from __future__ import annotations
import logging
from collections import defaultdict
from functools import cmp_to_key

import networkx
from claripy import ClaripyError

from angr import SIM_PROCEDURES
from angr import options as o
from angr.analyses import AnalysesHub
from angr.knowledge_base import KnowledgeBase
from angr.errors import AngrError, AngrCFGError, SimValueError, SimSolverModeError, SimError
from angr.sim_options import BYPASS_VERITESTING_EXCEPTIONS
from angr.sim_manager import SimulationManager
from angr.utils.graph import shallow_reverse
from . import Analysis, CFGEmulated

l = logging.getLogger(name=__name__)


class VeritestingError(Exception):
    pass


class CallTracingFilter:
    """
    Filter to apply during CFG creation on a given state and jumpkind to determine if it should be skipped at a certain
    depth
    """

    whitelist = {
        SIM_PROCEDURES["cgc"]["receive"],
        SIM_PROCEDURES["cgc"]["transmit"],
        SIM_PROCEDURES["posix"]["read"],
        SIM_PROCEDURES["libc"]["fgetc"],
        SIM_PROCEDURES["glibc"]["__ctype_b_loc"],
        SIM_PROCEDURES["libc"]["strlen"],
        SIM_PROCEDURES["libc"]["strcmp"],
        SIM_PROCEDURES["libc"]["atoi"],
    }

    cfg_cache = {}

    def __init__(self, project, depth, blacklist=None):
        self.project = project
        self.blacklist = [] if blacklist is None else blacklist
        self._skipped_targets = set()
        self.depth = depth

    def filter(self, call_target_state, jumpkind):
        """
        The call will be skipped if it returns True.

        :param call_target_state:   The new state of the call target.
        :param jumpkind:            The Jumpkind of this call.
        :returns:                   True if we want to skip this call, False otherwise.
        """

        ACCEPT = False
        REJECT = True

        l.debug("Filtering calling target %s", call_target_state.ip)

        # Currently we always skip the call, unless the target function satisfies one of the following conditions:
        # 1) It's a SimProcedure that are in the whitelist
        # 2) It's a function that has no loops, and no calls/syscalls,
        # 3) It's a function that has no loops, and only has calls to another function that will not be filtered out by
        #    this filter

        # Generate a CFG
        ip = call_target_state.ip

        if self.depth >= 5:
            l.debug("Rejecting target %s - too deep, depth is %d", ip, self.depth)
            return REJECT

        try:
            addr = call_target_state.solver.eval_one(ip)
        except (SimValueError, SimSolverModeError):
            self._skipped_targets.add(-1)
            l.debug("Rejecting target %s - cannot be concretized", ip)
            return REJECT

        # Is it in our blacklist?
        if addr in self.blacklist:
            self._skipped_targets.add(addr)
            l.debug("Rejecting target 0x%x - blacklisted", addr)
            return REJECT

        # If the target is a SimProcedure, is it on our whitelist?
        if self.project.is_hooked(addr) and type(self.project._sim_procedures[addr][0]) in CallTracingFilter.whitelist:
            # accept!
            l.debug("Accepting target 0x%x, jumpkind %s", addr, jumpkind)
            return ACCEPT

        # If it's a syscall, let's see if the real syscall is inside our whitelist
        if jumpkind.startswith("Ijk_Sys"):
            call_target_state.history.jumpkind = jumpkind
            successors_ = self.project.factory.successors(call_target_state)
            try:
                next_run = successors_.artifacts["procedure"]
            except KeyError:
                l.warning("CallTracingFilter.filter(): except artifacts['procedure'] in %s. Reject.", successors_)
                return REJECT

            if type(next_run) in CallTracingFilter.whitelist:
                # accept!
                l.debug("Accepting target 0x%x, jumpkind %s", addr, jumpkind)
                return ACCEPT
            # reject
            l.debug("Rejecting target 0x%x - syscall %s not in whitelist", addr, type(next_run))
            return REJECT

        cfg_key = (addr, jumpkind, self.project.filename)
        if cfg_key not in self.cfg_cache:
            new_blacklist = self.blacklist[::]
            new_blacklist.append(addr)
            tracing_filter = CallTracingFilter(self.project, depth=self.depth + 1, blacklist=new_blacklist)
            cfg = self.project.analyses[CFGEmulated].prep(kb=KnowledgeBase(self.project))(
                starts=((addr, jumpkind),),
                initial_state=call_target_state,
                context_sensitivity_level=0,
                call_depth=1,
                call_tracing_filter=tracing_filter.filter,
                normalize=True,
            )
            self.cfg_cache[cfg_key] = (cfg, tracing_filter)

            try:
                cfg.force_unroll_loops(1)
            except AngrCFGError:
                # Exceptions occurred during loop unrolling
                # reject
                l.debug("Rejecting target %#x - loop unrolling failed", addr)
                return REJECT

        else:
            l.debug("Loading CFG from CFG cache")
            cfg, tracing_filter = self.cfg_cache[cfg_key]

        if cfg._loop_back_edges:
            # It has loops!
            self._skipped_targets.add(addr)
            l.debug("Rejecting target 0x%x - it has loops", addr)
            return REJECT

        sim_procedures = [n for n in cfg.graph.nodes() if n.simprocedure_name is not None]
        for sp_node in sim_procedures:
            if not self.project.is_hooked(sp_node.addr):
                # This is probably a PathTerminator
                # Just skip it for now
                continue

            if self.project._sim_procedures[sp_node.addr].procedure not in CallTracingFilter.whitelist:
                self._skipped_targets.add(addr)
                l.debug("Rejecting target 0x%x - contains SimProcedures outside whitelist", addr)
                return REJECT

        if len(tracing_filter._skipped_targets):
            # Bummer
            self._skipped_targets.add(addr)
            l.debug("Rejecting target 0x%x - should be skipped", addr)
            return REJECT

        # accept!
        l.debug("Accepting target 0x%x, jumpkind %s", addr, jumpkind)
        return ACCEPT


class Veritesting(Analysis):
    """
    An exploration technique made for condensing chunks of code to single (nested) if-then-else constraints via CFG
    accurate to conduct Static Symbolic Execution SSE (conversion to single constraint)
    """

    # A cache for CFG we generated before
    cfg_cache = {}
    # Names of all stashes we will return from Veritesting
    all_stashes = ("successful", "errored", "deadended", "deviated", "unconstrained")

    def __init__(
        self,
        input_state,
        boundaries=None,
        loop_unrolling_limit=10,
        enable_function_inlining=False,
        terminator=None,
        deviation_filter=None,
    ):
        """
        SSE stands for Static Symbolic Execution, and we also implemented an extended version of Veritesting (Avgerinos,
        Thanassis, et al, ICSE 2014).

        :param input_state:              The initial state to begin the execution with.
        :param boundaries:               Addresses where execution should stop.
        :param loop_unrolling_limit:     The maximum times that Veritesting should unroll a loop for.
        :param enable_function_inlining: Whether we should enable function inlining and syscall inlining.
        :param terminator:               A callback function that takes a state as parameter. Veritesting will terminate
                                         if this function returns True.
        :param deviation_filter:         A callback function that takes a state as parameter. Veritesting will put the
                                         state into "deviated" stash if this function returns True.
        """
        block = self.project.factory.block(input_state.addr)
        branches = block.vex.constant_jump_targets_and_jumpkinds

        # if we are not at a conditional jump, just do a normal step
        if list(branches.values()) != ["Ijk_Boring", "Ijk_Boring"]:
            self.result, self.final_manager = False, None
            return

        # otherwise do a veritesting step

        self._input_state = input_state.copy()
        self._boundaries = boundaries if boundaries is not None else []
        self._loop_unrolling_limit = loop_unrolling_limit
        self._enable_function_inlining = enable_function_inlining
        self._terminator = terminator
        self._deviation_filter = deviation_filter

        # set up the cfg stuff
        self._cfg, self._loop_graph = self._make_cfg()
        self._loop_backedges = self._cfg._loop_back_edges
        self._loop_heads = {dst.addr for _, dst in self._loop_backedges}

        l.info("Static symbolic execution starts at %#x", self._input_state.addr)
        l.debug(
            "The execution will terminate at the following addresses: [ %s ]",
            ", ".join([hex(i) for i in self._boundaries]),
        )

        l.debug("A loop will be unrolled by a maximum of %d times.", self._loop_unrolling_limit)
        if self._enable_function_inlining:
            l.debug("Function inlining is enabled.")
        else:
            l.debug("Function inlining is disabled.")

        self.result, self.final_manager = self._veritesting()

    def _veritesting(self) -> tuple[bool, SimulationManager]:
        """
        Perform static symbolic execution starting from the given point.
        :returns: tuple of the success/failure of veritesting and the subsequent SimulationManager after execution
        """

        s = self._input_state.copy()

        try:
            new_manager = self._execute_and_merge(s)

        except (ClaripyError, SimError, AngrError):
            if BYPASS_VERITESTING_EXCEPTIONS not in s.options:
                raise
            l.warning("Veritesting caught an exception.", exc_info=True)
            return False, SimulationManager(self.project, stashes={"deviated": [s]})

        except VeritestingError as ex:
            l.warning("Exception occurred: %s", str(ex))
            return False, SimulationManager(self.project, stashes={"deviated": [s]})

        l.info(
            "Returning new paths: (successful: %s, deadended: %s, errored: %s, deviated: %s)",
            len(new_manager.successful),
            len(new_manager.deadended),
            len(new_manager.errored),
            len(new_manager.deviated),
        )

        return True, new_manager

    def _execute_and_merge(self, state):
        """
        Symbolically execute the program in a static manner. The basic idea is that we look ahead by creating a CFG,
        then perform a _controlled symbolic exploration_ based on the CFG, one path at a time. The controlled symbolic
        exploration stops when it sees a branch whose both directions are all feasible, or it shall wait for a merge
        from another path.

        A basic block will not be executed for more than *loop_unrolling_limit* times. If that is the case, a new state
        will be returned.

        :param SimState state: The initial state to start the execution.
        :returns:         A list of new states.
        """

        # Find all merge points
        merge_points = self._get_all_merge_points(self._cfg, self._loop_graph)
        l.debug("Merge points: %s", [hex(i[0]) for i in merge_points])

        #
        # Controlled symbolic exploration
        #

        # Initialize the beginning state
        initial_state = state
        initial_state.globals["loop_ctrs"] = defaultdict(int)

        manager = SimulationManager(
            self.project,
            active_states=[initial_state],
            resilience=o.BYPASS_VERITESTING_EXCEPTIONS in initial_state.options,
        )

        # Initialize all stashes
        for stash in self.all_stashes:
            manager.stashes[stash] = []
        # immediate_dominators = cfg.immediate_dominators(cfg.get_any_node(ip_int))

        while manager.active:
            # Step one step forward
            l.debug("Steps %s with %d active states: [ %s ]", manager, len(manager.active), manager.active)

            # Apply self.deviation_func on every single active state, and move them to deviated stash if needed
            if self._deviation_filter is not None:
                manager.stash(filter_func=self._deviation_filter, from_stash="active", to_stash="deviated")

            # Mark all those paths that are out of boundaries as successful
            manager.stash(filter_func=self.is_overbound, from_stash="active", to_stash="successful")

            manager.step(successor_func=self._get_successors)

            if self._terminator is not None and self._terminator(manager):
                for p in manager.stashes[stash]:
                    self._unfuck(p)
                break

            # Stash all paths that we do not see in our CFG
            manager.stash(filter_func=self.is_not_in_cfg, to_stash="deviated")

            # Stash all paths that we do not care about
            manager.stash(
                filter_func=lambda state: (
                    state.history.jumpkind not in ("Ijk_Boring", "Ijk_Call", "Ijk_Ret", "Ijk_NoHook")
                    and not state.history.jumpkind.startswith("Ijk_Sys")
                ),
                to_stash="deadended",
            )

            if manager.deadended:
                l.debug("Now we have some deadended paths: %s", manager.deadended)

            # Stash all possible states that we should merge later
            for merge_point_addr, merge_point_looping_times in merge_points:
                manager.stash(
                    lambda s, merge_point_addr=merge_point_addr: s.addr == merge_point_addr,
                    to_stash=f"_merge_{merge_point_addr:x}_{merge_point_looping_times}",
                )

            # Try to merge a set of previously stashed paths, and then unstash them
            if not manager.active:
                manager = self._join_merge_points(manager, merge_points)
        if any(len(manager.stashes[stash_name]) for stash_name in self.all_stashes):
            # Remove all stashes other than errored or deadended
            for stash in list(manager.stashes):
                if stash not in self.all_stashes:
                    manager.drop(stash=stash)

            for stash in manager.stashes:
                manager.apply(self._unfuck, stash=stash)

        return manager

    def _join_merge_points(self, manager, merge_points):
        """
        Merges together the appropriate execution points and unstashes them from the intermediate merge_x_y stashes to
        pruned (dropped), deadend or active stashes

        param SimulationManager manager:        current simulation context being stepped through
        param [(int, int)] merge_points: list of address and loop counters of execution points to merge
        returns SimulationManager:              new manager with edited stashes
        """
        merged_anything = False
        for merge_point_addr, merge_point_looping_times in merge_points:
            if merged_anything:
                break

            stash_name = f"_merge_{merge_point_addr:x}_{merge_point_looping_times}"
            if stash_name not in manager.stashes:
                continue

            stash_size = len(manager.stashes[stash_name])
            if stash_size == 0:
                continue
            if stash_size == 1:
                l.info("Skipping merge of 1 state in stash %s.", stash_size)
                manager.move(stash_name, "active")
                continue

            # let everyone know of the impending disaster
            l.info("Merging %d states in stash %s", stash_size, stash_name)

            # Try to prune the stash, so unsatisfiable states will be thrown away
            manager.prune(from_stash=stash_name, to_stash="pruned")
            if "pruned" in manager.stashes and len(manager.pruned):
                l.debug("... pruned %d paths from stash %s", len(manager.pruned), stash_name)
            # Remove the pruned stash to save memory
            manager.drop(stash="pruned")

            # merge things callstack by callstack
            while len(manager.stashes[stash_name]):
                r = manager.stashes[stash_name][0]
                manager.move(stash_name, "merge_tmp", lambda p, r=r: p.callstack == r.callstack)

                old_count = len(manager.merge_tmp)
                l.debug("... trying to merge %d states.", old_count)

                # merge the loop_ctrs
                new_loop_ctrs = defaultdict(int)
                for m in manager.merge_tmp:
                    for head_addr, looping_times in m.globals["loop_ctrs"].items():
                        new_loop_ctrs[head_addr] = max(looping_times, m.globals["loop_ctrs"][head_addr])

                manager.merge(stash="merge_tmp")
                for m in manager.merge_tmp:
                    m.globals["loop_ctrs"] = new_loop_ctrs

                new_count = len(manager.stashes["merge_tmp"])
                l.debug("... after merge: %d states.", new_count)

                merged_anything |= new_count != old_count

                if len(manager.merge_tmp) > 1:
                    l.warning("More than 1 state after Veritesting merge.")
                    manager.move("merge_tmp", "active")
                elif any(
                    loop_ctr >= self._loop_unrolling_limit + 1
                    for loop_ctr in manager.one_merge_tmp.globals["loop_ctrs"].values()
                ):
                    l.debug("... merged state is overlooping")
                    manager.move("merge_tmp", "deadended")
                else:
                    l.debug("... merged state going to active stash")
                    manager.move("merge_tmp", "active")

        return manager

    #
    # Path management
    #

    def is_not_in_cfg(self, s):
        """
        Returns if s.addr is not a proper node in our CFG.

        :param SimState s: The SimState instance to test.
        :returns bool: False if our CFG contains p.addr, True otherwise.
        """

        n = self._cfg.model.get_any_node(s.addr, is_syscall=s.history.jumpkind.startswith("Ijk_Sys"))
        if n is None:
            return True

        return n.simprocedure_name == "PathTerminator"

    def _get_successors(self, state):
        """
        Gets the successors to the current state by step, saves copy of state and finally stashes new unconstrained
        states to manager.

        :param SimState state:          Current state to step on from
        :returns SimSuccessors:         The SimSuccessors object
        """
        size_of_next_irsb = self._cfg.model.get_any_node(state.addr).size
        return self.project.factory.successors(state, size=size_of_next_irsb)

    def is_overbound(self, state):
        """
        Filter out all states that run out of boundaries or loop too many times.

        param SimState state: SimState instance to check
        returns bool:    True if outside of mem/loop_ctr boundary
        """

        ip = state.addr

        if ip in self._boundaries:
            l.debug("... terminating Veritesting due to overbound")
            return True

        try:
            # If the address is not in the list (which could mean it is
            # not at the top of a block), check directly in the blocks
            # (Blocks are repeatedly created for every check, but with
            # the IRSB cache in angr lifter it should be OK.)
            if set(self._boundaries).intersection(set(self.project.factory.block(ip).instruction_addrs)):
                return True
        except (AngrError, SimError):
            pass

        if (
            ip in self._loop_heads  # This is the beginning of the loop
            or state.history.jumpkind == "Ijk_Call"  # We also wanna catch recursive function calls
        ):
            state.globals["loop_ctrs"][ip] += 1
            if state.globals["loop_ctrs"][ip] >= self._loop_unrolling_limit + 1:
                l.debug("... terminating Veritesting due to overlooping")
                return True

        l.debug("... accepted")
        return False

    @staticmethod
    def _unfuck(s):
        """
        Deletes the loop counter from state's information dictionary

        :param SimState s: SimState instance to update
        :returns SimState: same SimState with deleted loop counter
        """
        del s.globals["loop_ctrs"]
        return s

    #
    # Merge point determination
    #

    def _make_cfg(self):
        """
        Builds a CFG from the current function.
        Saved in cfg_cache.

        returns (CFGEmulated, networkx.DiGraph): Tuple of the CFG and networkx representation of it
        """

        state = self._input_state
        ip_int = state.addr

        # if the cfg is cached, simply return the cached cfg
        cfg_key = (ip_int, state.history.jumpkind, self.project.filename)
        if cfg_key in self.cfg_cache:
            cfg, cfg_graph_with_loops = self.cfg_cache[cfg_key]
            return cfg, cfg_graph_with_loops

        if self._enable_function_inlining:
            call_tracing_filter = CallTracingFilter(self.project, depth=0)
            filter = call_tracing_filter.filter  # pylint:disable=redefined-builtin
        else:
            filter = None

        # To better handle syscalls, we make a copy of all registers if they are not symbolic
        cfg_initial_state = self.project.factory.blank_state(mode="fastpath")

        # FIXME: This is very hackish
        # FIXME: And now only Linux-like syscalls are supported
        if self.project.arch.name == "X86":
            if not state.solver.symbolic(state.regs.eax):
                cfg_initial_state.regs.eax = state.regs.eax
        elif self.project.arch.name == "AMD64" and not state.solver.symbolic(state.regs.rax):
            cfg_initial_state.regs.rax = state.regs.rax

        # generate the cfg and perform loop unrolling
        cfg = self.project.analyses[CFGEmulated].prep(kb=KnowledgeBase(self.project))(
            starts=((ip_int, state.history.jumpkind),),
            context_sensitivity_level=0,
            call_depth=1,
            call_tracing_filter=filter,
            initial_state=cfg_initial_state,
            normalize=True,
        )
        cfg_graph_with_loops = networkx.DiGraph(cfg.graph)
        cfg.force_unroll_loops(self._loop_unrolling_limit)

        # cache the generated cfg
        self.cfg_cache[cfg_key] = (cfg, cfg_graph_with_loops)

        return cfg, cfg_graph_with_loops

    @staticmethod
    def _post_dominate(reversed_graph, n1, n2):
        """
        Checks whether `n1` post-dominates `n2` in the *original* (not reversed) graph.

        :param networkx.DiGraph reversed_graph:  The reversed networkx.DiGraph instance.
        :param networkx.Node n1:                 Node 1.
        :param networkx.Node n2:                 Node 2.
        :returns bool:                           True/False.
        """

        ds = networkx.immediate_dominators(reversed_graph, n1)
        return n2 in ds

    def _get_all_merge_points(self, cfg, graph_with_loops):
        """
        Return all possible merge points in this CFG.

        :param CFGEmulated cfg: The control flow graph, which must be acyclic.
        :returns [(int, int)]:  A list of merge points (address and number of times looped).
        """

        graph = networkx.DiGraph(cfg.graph)
        reversed_cyclic_graph = shallow_reverse(graph_with_loops)

        # Remove all "FakeRet" edges
        fakeret_edges = [
            (src, dst) for src, dst, data in graph.edges(data=True) if data["jumpkind"] in ("Ijk_FakeRet", "Ijk_Exit")
        ]
        graph.remove_edges_from(fakeret_edges)

        # Remove all "FakeRet" edges from cyclic_graph as well
        fakeret_edges = [
            (src, dst)
            for src, dst, data in reversed_cyclic_graph.edges(data=True)
            if data["jumpkind"] in ("Ijk_FakeRet", "Ijk_Exit")
        ]
        reversed_cyclic_graph.remove_edges_from(fakeret_edges)

        # Perform a topological sort
        sorted_nodes = networkx.topological_sort(graph)

        nodes = [n for n in sorted_nodes if graph.in_degree(n) > 1 and n.looping_times == 0]

        # Reorder nodes based on post-dominance relations
        nodes = sorted(
            nodes,
            key=cmp_to_key(
                lambda n1, n2: (
                    1
                    if self._post_dominate(reversed_cyclic_graph, n1, n2)
                    else (-1 if self._post_dominate(reversed_cyclic_graph, n2, n1) else 0)
                )
            ),
        )

        return [(n.addr, n.looping_times) for n in nodes]


AnalysesHub.register_default("Veritesting", Veritesting)



from __future__ import annotations
from .plugin import KnowledgeBasePlugin


class Comments(KnowledgeBasePlugin, dict):
    """
    Tracks comments via a Dict of Address -> Text
    """

    def copy(self):
        o = Comments(self._kb)
        o.update(self)
        return o


KnowledgeBasePlugin.register_default("comments", Comments)



from __future__ import annotations


class XRefType:
    Offset = 0
    Read = 1
    Write = 2

    @staticmethod
    def to_string(ty):
        s = {
            XRefType.Offset: "offset",
            XRefType.Read: "read",
            XRefType.Write: "write",
        }
        return s.get(ty, "unknown")



from __future__ import annotations

from .region_identifier import RegionIdentifier
from .structured_codegen import CStructuredCodeGenerator, ImportSourceCode
from .clinic import Clinic
from .region_simplifiers import RegionSimplifier
from .decompiler import Decompiler
from .decompilation_options import options, options_by_category
from .block_simplifier import BlockSimplifier
from .callsite_maker import CallSiteMaker
from .ail_simplifier import AILSimplifier
from .ssailification import Ssailification
from .dephication import GraphDephication, SeqNodeDephication
from .presets import DECOMPILATION_PRESETS
from . import structuring
from . import optimization_passes


StructuredCodeGenerator = CStructuredCodeGenerator


__all__ = (
    "DECOMPILATION_PRESETS",
    "AILSimplifier",
    "BlockSimplifier",
    "CStructuredCodeGenerator",
    "CallSiteMaker",
    "Clinic",
    "Decompiler",
    "GraphDephication",
    "ImportSourceCode",
    "RegionIdentifier",
    "RegionSimplifier",
    "SeqNodeDephication",
    "Ssailification",
    "StructuredCodeGenerator",
    "optimization_passes",
    "options",
    "options_by_category",
    "structuring",
)



from __future__ import annotations


class SAILRSemanticError(Exception):
    """
    These types of errors may not kill the entire analysis, but they do kill the current working round.
    They are caused when some non-expected state happens, like a ConditionalJump that is not the last statement in a
    block.
    """


class UnsupportedAILNodeError(SAILRSemanticError):
    """
    These type of errors are raised when an AIL node is not supported by the current implementation.
    Most common with Switch heads.
    """



from __future__ import annotations
import math

import claripy

# Copied from engines/vex/claripy/irop.py
fp_rm_map = {
    0: claripy.fp.RM.RM_NearestTiesEven,
    1: claripy.fp.RM.RM_TowardsNegativeInf,
    2: claripy.fp.RM.RM_TowardsPositiveInf,
    3: claripy.fp.RM.RM_TowardsZero,
}


def translate_rm(rm_num):
    if not rm_num.symbolic:
        return fp_rm_map[rm_num.concrete_value]

    return claripy.fp.RM.default()


def concretize_2xm1(state, args):
    # 2xm1(x) = 2 ** x - 1. Concretize 2**x part alone since only that cannot be modelled in Z3.
    arg_x = state.solver.eval(args[1])
    if -1 <= arg_x <= 1:
        return claripy.FPV(math.pow(2, arg_x) - 1, claripy.FSORT_DOUBLE)

    # If x is outside range [-1.0, 1.0], result is undefined. We return argument itself as observed on an Intel CPU.
    return args[1]


def concretize_abs_float64(state, args):
    arg_val = state.solver.eval(args[0])
    return claripy.FPV(abs(arg_val), args[0].sort)


def concretize_add_float64(state, args):
    arg0 = state.solver.eval(args[1])
    arg1 = state.solver.eval(args[2])
    return claripy.FPV(arg0 + arg1, claripy.FSORT_DOUBLE)


def concretize_add32f04(state, args):
    fp_arg0 = state.solver.eval(args[0][31:0].raw_to_fp())
    fp_arg1 = state.solver.eval(args[1][31:0].raw_to_fp())
    result = claripy.FPV(fp_arg0 + fp_arg1, claripy.FSORT_FLOAT).raw_to_bv()
    return claripy.Concat(args[0][(args[0].length - 1) : result.size()], result)


def concretize_add64f02(state, args):
    fp_arg0 = state.solver.eval(args[0][63:0].raw_to_fp())
    fp_arg1 = state.solver.eval(args[1][63:0].raw_to_fp())
    result = claripy.FPV(fp_arg0 + fp_arg1, claripy.FSORT_DOUBLE).raw_to_bv()
    return claripy.Concat(args[0][(args[0].length - 1) : result.size()], result)


def concretize_cmpf64(state, args):
    fp_arg0 = state.solver.eval(args[0])
    fp_arg1 = state.solver.eval(args[1])
    if fp_arg0 < fp_arg1:
        return claripy.BVV(1, 32)
    if fp_arg0 > fp_arg1:
        return claripy.BVV(0, 32)
    if fp_arg0 == fp_arg1:
        return claripy.BVV(0x40, 32)

    return claripy.BVV(0x45, 32)


def concretize_divf64(state, args):
    arg1 = state.solver.eval(args[1])
    arg2 = state.solver.eval(args[2])
    if arg2 == 0:
        return claripy.FPV(math.inf, args[1].sort)

    return claripy.FPV(arg1 / arg2, args[1].sort)


def concretize_div32f04(state, args):
    fp_arg0 = state.solver.eval(args[0][31:0].raw_to_fp())
    fp_arg1 = state.solver.eval(args[1][31:0].raw_to_fp())
    if fp_arg1 == 0:
        result = claripy.FPV(math.inf, claripy.FSORT_FLOAT).raw_to_bv()
    else:
        result = claripy.FPV(fp_arg0 / fp_arg1, claripy.FSORT_FLOAT).raw_to_bv()

    return claripy.Concat(args[0][(args[0].length - 1) : result.size()], result)


def concretize_div64f02(state, args):
    fp_arg0 = state.solver.eval(args[0][63:0].raw_to_fp())
    fp_arg1 = state.solver.eval(args[1][63:0].raw_to_fp())
    if fp_arg1 == 0:
        result = claripy.FPV(math.inf, claripy.FSORT_DOUBLE).raw_to_bv()
    else:
        result = claripy.FPV(fp_arg0 / fp_arg1, claripy.FSORT_DOUBLE).raw_to_bv()

    return claripy.Concat(args[0][(args[0].length - 1) : result.size()], result)


def concretize_float32_to_float64(state, args):
    arg0 = state.solver.eval(args[0])
    return claripy.FPV(arg0, claripy.FSORT_DOUBLE)


def concretize_float64_to_float32(state, args):
    arg = state.solver.eval(args[1])
    return claripy.FPV(arg, claripy.FSORT_FLOAT)


def concretize_float64_to_int64s(state, args):
    rm = translate_rm(args[0])
    return state.solver.fpToSBV(rm, args[1], 64)


def concretize_int32s_to_float64(state, args):
    arg = claripy.BVV(state.solver.eval(args[0]), args[0].size())
    return arg.val_to_fp(claripy.fp.FSort.from_size(64), signed=True, rm=fp_rm_map[0])


def concretize_int64s_to_float64(state, args):
    rm = translate_rm(args[0])
    arg = claripy.BVV(state.solver.eval(args[1]), args[1].size())
    return arg.val_to_fp(claripy.fp.FSort.from_size(64), signed=True, rm=rm)


def concretize_mulf64(state, args):
    arg1 = state.solver.eval(args[1])
    arg2 = state.solver.eval(args[2])
    return claripy.FPV(arg1 / arg2, args[1].sort)


def concretize_mul32f04(state, args):
    fp_arg0 = state.solver.eval(args[0][31:0].raw_to_fp())
    fp_arg1 = state.solver.eval(args[1][31:0].raw_to_fp())
    result = claripy.FPV(fp_arg0 * fp_arg1, claripy.FSORT_FLOAT).raw_to_bv()
    return claripy.Concat(args[0][(args[0].length - 1) : result.size()], result)


def concretize_mul64f02(state, args):
    fp_arg0 = state.solver.eval(args[0][63:0].raw_to_fp())
    fp_arg1 = state.solver.eval(args[1][63:0].raw_to_fp())
    result = claripy.FPV(fp_arg0 * fp_arg1, claripy.FSORT_DOUBLE).raw_to_bv()
    return claripy.Concat(args[0][(args[0].length - 1) : result.size()], result)


def concretize_fscale(state, args):
    # fscale(x, y) = x * (2 ** y)
    arg_x = state.solver.eval(args[1])
    arg_y = math.floor(state.solver.eval(args[2]))
    if math.isnan(arg_x) or math.isnan(arg_y):
        return claripy.FPV(math.nan, claripy.FSORT_DOUBLE)

    if abs(arg_x) == math.inf and arg_y == -1 * math.inf:
        return claripy.FPV(math.nan, claripy.FSORT_DOUBLE)

    if arg_x == 0.0 and arg_y == math.inf:
        return claripy.FPV(math.nan, claripy.FSORT_DOUBLE)

    return claripy.FPV(arg_x * math.pow(2, arg_y), claripy.FSORT_DOUBLE)


def concretize_fsqrt32(state, args):
    # Concretize floating point square root. Z3 does support square root but unsure if that includes floating point
    arg_1 = state.solver.eval(args[1])
    if arg_1 < 0 or math.isnan(arg_1):
        return claripy.FPV(math.nan, claripy.FSORT_FLOAT)

    return claripy.FPV(math.sqrt(arg_1), claripy.FSORT_FLOAT)


def concretize_fsqrt(state, args):
    # Concretize floating point square root. Z3 does support square root but unsure if that includes floating point
    arg_1 = state.solver.eval(args[1])
    if arg_1 < 0 or math.isnan(arg_1):
        return claripy.FPV(math.nan, claripy.FSORT_DOUBLE)

    return claripy.FPV(math.sqrt(arg_1), claripy.FSORT_DOUBLE)


def concretize_prem(state, args):
    # Compute partial remainder. Z3 does not support modulo for reals: https://github.com/Z3Prover/z3/issues/557
    # Implementation based on description in the Intel software manual
    dividend = state.solver.eval(args[1])
    divisor = state.solver.eval(args[2])
    if math.isnan(dividend) or math.isnan(divisor) or abs(dividend) == math.inf or divisor == 0.0:
        return claripy.FPV(math.nan, claripy.FSORT_DOUBLE)

    if abs(divisor) == math.inf or dividend == 0.0:
        return args[1]

    _, exp_dividend = math.frexp(dividend)
    _, exp_divisor = math.frexp(divisor)
    if exp_dividend - exp_divisor < 64:
        quotient = math.floor(dividend / divisor)
        result = dividend - (divisor * quotient)
    else:
        # According to Intel manual, N is an implementation-dependent number between 32 and 63. 60 was chosen
        # arbitrarily.
        N = 60
        quotient = math.floor((dividend / divisor) / pow(2, exp_dividend - exp_divisor - N))
        result = dividend - (divisor * quotient * pow(2, exp_dividend - exp_divisor - N))

    if result == 0.0 and math.copysign(1.0, dividend) < 0:
        # According to Intel manual, if result is 0, its sign should be same as that of dividend.
        return claripy.FPV(-0.0, claripy.FSORT_DOUBLE)

    return claripy.FPV(result, claripy.FSORT_DOUBLE)


def concretize_prem_flags(state, args):
    # Compute FPU flags for partial remainder.
    # Implementation based on description in the Intel software manual
    dividend = state.solver.eval(args[1])
    divisor = state.solver.eval(args[2])
    # pylint: disable=too-many-boolean-expressions
    if (
        math.isnan(dividend)
        or math.isnan(divisor)
        or abs(dividend) == math.inf
        or divisor == 0.0
        or abs(divisor) == math.inf
        or dividend == 0.0
    ):
        # Since these are exception cases, the manual does not specify anything for these flags. These are set to
        # zero based on what was observed on an Intel CPU.
        flag_c0 = 0
        flag_c1 = 0
        flag_c2 = 0
        flag_c3 = 0
    else:
        _, exp_dividend = math.frexp(dividend)
        _, exp_divisor = math.frexp(divisor)
        if exp_dividend - exp_divisor < 64:
            quotient = math.floor(dividend / divisor)
            flag_c2 = 0
            flag_c0 = (quotient & 4) >> 2
            flag_c3 = (quotient & 2) >> 1
            flag_c1 = quotient & 1
        else:
            # Nothing is explicitly specified for C0 and C3 bits in this case so arbitrarily set to 0
            # TODO: C1 should be set to 0 only if floating point stack underflows. How to detect that?
            flag_c2 = 1
            flag_c0 = 0
            flag_c1 = 0
            flag_c3 = 0

    flags = (flag_c3 << 14) | (flag_c2 << 10) | (flag_c1 << 9) | (flag_c0 << 8)
    return claripy.BVV(flags, 16)


def concretize_reinterp_float64_as_int64(state, args):
    return claripy.FPV(state.solver.eval(args[0]), args[0].sort).raw_to_bv()


def concretize_sub32f04(state, args):
    fp_arg0 = state.solver.eval(args[0][31:0].raw_to_fp())
    fp_arg1 = state.solver.eval(args[1][31:0].raw_to_fp())
    result = claripy.FPV(fp_arg0 - fp_arg1, claripy.FSORT_FLOAT).raw_to_bv()
    return claripy.Concat(args[0][(args[0].length - 1) : result.size()], result)


def concretize_sub64f02(state, args):
    fp_arg0 = state.solver.eval(args[0][63:0].raw_to_fp())
    fp_arg1 = state.solver.eval(args[1][63:0].raw_to_fp())
    result = claripy.FPV(fp_arg0 - fp_arg1, claripy.FSORT_DOUBLE).raw_to_bv()
    return claripy.Concat(args[0][(args[0].length - 1) : result.size()], result)


def concretize_trig_cos(state, args):
    # cos(x). Z3 does support *some* cases of cos(see https://github.com/Z3Prover/z3/issues/680) but we don't use
    # the feature and concretize fully instead.
    arg_x = state.solver.eval(args[1])
    abs_arg_x = abs(arg_x)

    if math.isnan(abs_arg_x):
        return claripy.FPV(math.nan, claripy.FSORT_DOUBLE)

    if abs_arg_x == math.inf:
        # Floating-point invalid-operation exception
        return claripy.FPV(math.nan, claripy.FSORT_DOUBLE)

    if abs_arg_x > pow(2, 63):
        # Intel manual says argument must be in range [-2^63, 2^63]. Otherwise, floating-point invalid-operation
        # exception: leave value changed
        return args[1]

    return claripy.FPV(math.cos(arg_x), claripy.FSORT_DOUBLE)


def concretize_trig_sin(state, args):
    # sin(x). Z3 does support *some* cases of sin(see https://github.com/Z3Prover/z3/issues/680) but we don't use
    # the feature and concretize fully instead.
    arg_x = state.solver.eval(args[1])
    abs_arg_x = abs(arg_x)

    if math.isnan(abs_arg_x):
        return claripy.FPV(math.nan, claripy.FSORT_DOUBLE)

    if abs_arg_x == math.inf:
        # Floating-point invalid-operation exception
        return claripy.FPV(math.nan, claripy.FSORT_DOUBLE)

    if abs_arg_x > pow(2, 63):
        # Intel manual says argument must be in range [-2^63, 2^63]. Otherwise, floating-point invalid-operation
        # exception: leave value changed
        return args[1]

    return claripy.FPV(math.sin(arg_x), claripy.FSORT_DOUBLE)


def concretize_trig_tan(state, args):
    # tan(x). Concretize fully since it cannot be modelled in Z3.
    # TODO: How to handle NaN arg?
    arg_x = state.solver.eval(args[1])
    abs_arg_x = abs(arg_x)

    if math.isnan(abs_arg_x):
        return claripy.FPV(math.nan, claripy.FSORT_DOUBLE)

    if abs_arg_x == math.inf:
        # Floating-point invalid-operation exception
        return claripy.FPV(math.nan, claripy.FSORT_DOUBLE)

    if abs_arg_x > pow(2, 63):
        # Intel manual says argument must be in range [-2^63, 2^63]. Otherwise, floating-point invalid-operation
        # exception: leave value changed
        return args[1]

    return claripy.FPV(math.tan(arg_x), claripy.FSORT_DOUBLE)


def concretize_yl2x(state, args):
    # yl2x(y, x) = y * log2(x)
    # 3 arguments are passed: first is FP rounding mode. y is from st(1) and x from st(0).
    # TODO: Return NaN if either arg is non-numeric.
    # TODO: Set FPU flags.
    arg_x = state.solver.eval(args[2])
    arg_y = state.solver.eval(args[1])
    if arg_x < 0:
        # TODO: Indicate floating-point invalid-operation exception
        return claripy.FPV(arg_x, claripy.FSORT_DOUBLE)

    if arg_x == 0:
        if abs(arg_y) == math.inf:
            return claripy.FPV(-1 * arg_y, claripy.FSORT_DOUBLE)
        if arg_y == 0:
            # TODO: Indicate floating-point invalid-operation exception
            return claripy.FPV(arg_x, claripy.FSORT_DOUBLE)
        # TODO: Indicate floating-point zero-division exception
        return claripy.FPV(arg_x, claripy.FSORT_DOUBLE)

    if arg_x == 1:
        if abs(arg_y) == math.inf:
            # TODO: Indicate floating-point invalid-operation exception
            return claripy.FPV(arg_x, claripy.FSORT_DOUBLE)

        # TODO: How to distinguish between +0 and -0?
        return claripy.FPV(0, claripy.FSORT_DOUBLE)

    if arg_x == math.inf:
        if arg_y == 0:
            # TODO: Indicate floating-point invalid-operation exception
            return claripy.FPV(arg_x, claripy.FSORT_DOUBLE)
        if arg_y < 0:
            return claripy.FPV(-1 * math.inf, claripy.FSORT_DOUBLE)

        return claripy.FPV(math.inf, claripy.FSORT_DOUBLE)

    return claripy.FPV(arg_y * math.log2(arg_x), claripy.FSORT_DOUBLE)


concretizers = {
    "Iop_Yl2xF64": concretize_yl2x,
    "Iop_ScaleF64": concretize_fscale,
    "Iop_2xm1F64": concretize_2xm1,
    "Iop_SqrtF32": concretize_fsqrt32,
    "Iop_SqrtF64": concretize_fsqrt,
    "Iop_CosF64": concretize_trig_cos,
    "Iop_SinF64": concretize_trig_sin,
    "Iop_TanF64": concretize_trig_tan,
    "Iop_PRemF64": concretize_prem,
    "Iop_PRemC3210F64": concretize_prem_flags,
    "Iop_I32StoF64": concretize_int32s_to_float64,
    "Iop_Mul64F0x2": concretize_mul64f02,
    "Iop_Add64F0x2": concretize_add64f02,
    "Iop_Div64F0x2": concretize_div64f02,
    "Iop_I64StoF64": concretize_int64s_to_float64,
    "Iop_DivF64": concretize_divf64,
    "Iop_MulF64": concretize_mulf64,
    "Iop_F64toI64S": concretize_float64_to_int64s,
    "Iop_AbsF64": concretize_abs_float64,
    "Iop_Sub64F0x2": concretize_sub64f02,
    "Iop_AddF64": concretize_add_float64,
    "Iop_F64toF32": concretize_float64_to_float32,
    "Iop_Div32F0x4": concretize_div32f04,
    "Iop_Sub32F0x4": concretize_sub32f04,
    "Iop_Add32F0x4": concretize_add32f04,
    "Iop_Mul32F0x4": concretize_mul32f04,
    "Iop_CmpF64": concretize_cmpf64,
    "Iop_F32toF64": concretize_float32_to_float64,
    "Iop_ReinterpF64asI64": concretize_reinterp_float64_as_int64,
}



from __future__ import annotations
from typing import TYPE_CHECKING

from collections.abc import Callable

from angr.engines.ail import ail_call_state
from angr.errors import AngrCallableError, AngrCallableMultistateError
from .callable import Callable as VEXCallable

if TYPE_CHECKING:
    from angr.analyses.decompiler.clinic import Clinic
    from angr.project import Project
    from angr.ailment import Address
    from angr.sim_manager import SimulationManager


class AILCallable(VEXCallable):
    """
    A Callable that uses AILCallState for its states.
    """

    def __init__(
        self,
        project: Project,
        addr: int | Address | str,
        lifter: Callable[[int], Clinic],
        mode: str = "symbolic",
        add_options: set[str] | None = None,
        remove_options: set[str] | None = None,
        preset: str = "default",
        boundary: set[int | Address] | None = None,
    ):
        super().__init__(project, addr, add_options=add_options, remove_options=remove_options)
        self._mode = mode
        self._lifter = lifter
        self._preset = preset
        self._boundary = boundary

    def perform_call(self, *args, prototype=None):
        state = ail_call_state(
            self._project,
            self._addr,
            args,
            self._lifter,
            mode=self._mode,
            add_options=self._add_options,
            remove_options=self._remove_options,
            plugin_preset=self._preset,
        )

        caller = self._project.factory.simulation_manager(state)
        caller.run(step_func=self._step_func).unstash(from_stash="deadended")
        caller.prune(filter_func=lambda pt: pt.addr == self._deadend_addr)

        if "step_limited" in caller.stashes:
            caller.stash(from_stash="step_limited", to_stash="active")
        if len(caller.active) == 0:
            raise AngrCallableError("No paths returned from function")

        self.result_path_group = caller.copy()

        if self._perform_merge:
            caller.merge()
            self.result_state = caller.active[0]
        elif len(caller.active) == 1:
            self.result_state = caller.active[0]

    def _step_func(self, pg: SimulationManager):
        pg2 = pg.prune()
        if self._concrete_only and len(pg2.active) > 1:
            raise AngrCallableMultistateError("Execution split on symbolic condition!")
        if self._step_limit:
            pg2.stash(filter_func=lambda p: p.history.depth >= self._step_limit, to_stash="step_limited")
        if self._boundary:
            pg2.stash(
                filter_func=lambda p: p.addr in self._boundary,
                to_stash="deadended",
            )
        return pg2



# pylint:disable=no-member,import-outside-toplevel
from __future__ import annotations

import contextlib
import logging

from angr import vaults
from .base import ExplorationTechnique


l = logging.getLogger(name=__name__)


class PickledStatesBase:
    """
    The base class of pickled states
    """

    def sort(self):
        """
        Sort pickled states.
        """

        raise NotImplementedError

    def add(self, prio, sid):
        """
        Add a newly pickled state.

        :param int prio:    Priority of the state.
        :param str sid:     Persistent ID of the state.
        :return:            None
        """
        raise NotImplementedError

    def pop_n(self, n):
        """
        Pop the top N states.

        :param int n:   Number of states to take.
        :return:        A list of states.
        """
        raise NotImplementedError


class PickledStatesList(PickledStatesBase):
    """
    List-backed pickled state storage.
    """

    def __init__(self):
        self._picked_states = []

    def sort(self):
        self._picked_states.sort()

    def add(self, prio, sid):
        self._picked_states.append((prio, sid))

    def pop_n(self, n):
        ss = self._picked_states[:n]
        self._picked_states[:n] = []
        return ss


class PickledStatesDb(PickledStatesBase):
    """
    Database-backed pickled state storage.
    """

    def __init__(self, db_str="sqlite:///:memory:"):
        from .spiller_db import sqlalchemy, create_engine, Base, OperationalError, sessionmaker

        if sqlalchemy is None:
            raise ImportError(
                f"Cannot import SQLAlchemy. Please install SQLAlchemy before using " f"{self.__class__.__name__}."
            )

        # ORM declarations
        engine = create_engine(db_str)

        # create table (unless it already exists)
        with contextlib.suppress(OperationalError):
            Base.metadata.create_all(engine, checkfirst=True)

        self.Session = sessionmaker(bind=engine)

    def sort(self):
        pass

    def add(self, prio, sid, taken=False, stash="spilled"):  # pylint:disable=arguments-differ
        from .spiller_db import PickledState

        record = PickledState(id=sid, priority=prio, taken=taken, stash=stash)
        session = self.Session()
        session.add(record)
        session.commit()
        session.close()

    def pop_n(self, n, stash="spilled"):  # pylint:disable=arguments-differ
        from .spiller_db import PickledState

        session = self.Session()
        q = (
            session.query(PickledState)
            .filter_by(taken=False)
            .filter_by(stash=stash)
            .order_by(PickledState.priority)
            .limit(n)
            .all()
        )

        ss = []
        for r in q:
            r.taken = True
            ss.append((r.priority, r.id))
        session.commit()
        session.close()
        return ss

    def get_recent_n(self, n, stash="spilled"):
        from .spiller_db import PickledState

        session = self.Session()
        q = session.query(PickledState).filter_by(stash=stash).order_by(PickledState.timestamp.desc()).limit(n).all()

        ss = []
        for r in q:
            ss.append((r.timestamp, r.id))
        session.close()
        return ss

    def count(self):
        from .spiller_db import PickledState

        session = self.Session()
        q = session.query(PickledState).count()
        session.close()
        return q


class Spiller(ExplorationTechnique):
    """
    Automatically spill states out. It can spill out states to a different stash, spill
    them out to ANA, or first do the former and then (after enough states) the latter.
    """

    def __init__(
        self,
        src_stash="active",
        min=5,
        max=10,  # pylint:disable=redefined-builtin
        staging_stash="spill_stage",
        staging_min=10,
        staging_max=20,
        pickle_callback=None,
        unpickle_callback=None,
        post_pickle_callback=None,
        priority_key=None,
        vault=None,
        states_collection=None,
    ):
        """
        Initializes the spiller.

        :param max:          the number of states that are *not* spilled
        :param src_stash:    the stash from which to spill states (default: active)
        :param staging_stash: the stash *to* which to spill states (default: "spill_stage")
        :param staging_max:  the number of states that can be in the staging stash before things get spilled to ANA
                             (default: None. If staging_stash is set, then this means unlimited, and ANA will not be
                             used).
        :param priority_key: a function that takes a state and returns its numerical priority (MAX_INT is lowest
                             priority). By default, self.state_priority will be used, which prioritizes by object ID.
        :param vault:        an angr.Vault object to handle storing and loading of states. If not provided, an
                             angr.vaults.VaultShelf will be created with a temporary file.
        """
        super().__init__()
        self.max = max
        self.min = min
        self.src_stash = src_stash
        self.staging_stash = staging_stash
        self.staging_max = staging_max
        self.staging_min = staging_min

        # various callbacks
        self.priority_key = priority_key
        self.unpickle_callback = unpickle_callback
        self.pickle_callback = pickle_callback
        self.post_pickle_callback = post_pickle_callback

        # tracking of pickled stuff
        self._pickled_states = PickledStatesList() if states_collection is None else states_collection
        self._ever_pickled = 0
        self._ever_unpickled = 0
        self._vault = vaults.VaultShelf() if vault is None else vault

    def _unpickle(self, n):
        self._pickled_states.sort()
        unpickled = [(sid, self._load_state(sid)) for _, sid in self._pickled_states.pop_n(n)]
        self._ever_unpickled += len(unpickled)
        if self.unpickle_callback:
            for sid, u in unpickled:
                self.unpickle_callback(sid, u)
        return [u for _, u in unpickled]

    def _get_priority(self, state):
        return (self.priority_key or self.state_priority)(state)

    def _pickle(self, states):
        if self.pickle_callback:
            for s in states:
                self.pickle_callback(s)
        self._ever_pickled += len(states)
        for state in states:
            try:
                state_oid = self._store_state(state)
            except RecursionError:
                l.warning(
                    "Couldn't store the state because of a recursion error. This is most likely to be pickle's "
                    "fault. You may try to increase the recursion limit using sys.setrecursionlimit()."
                )
                continue
            prio = self._get_priority(state)
            if self.post_pickle_callback:
                self.post_pickle_callback(state, prio, state_oid)
            self._pickled_states.add(prio, state_oid)

    def _store_state(self, state):
        return self._vault.store(state)

    def _load_state(self, sid):
        return self._vault.load(sid)

    def step(self, simgr, stash="active", **kwargs):
        simgr = simgr.step(stash=stash, **kwargs)

        l.debug(
            "STASH STATUS: active: %d, staging: %d",
            len(simgr.stashes[self.src_stash]),
            len(simgr.stashes[self.staging_stash]),
        )

        states = simgr.stashes[self.src_stash]
        staged_states = simgr.stashes.setdefault(self.staging_stash, []) if self.staging_stash else []

        if len(states) < self.min:
            missing = (self.max + self.min) // 2 - len(states)
            l.debug("Too few states (%d/%d) in stash %s.", len(states), self.min, self.src_stash)
            if self.staging_stash:
                l.debug("... retrieving states from staging stash (%s)", self.staging_stash)
                staged_states.sort(key=self.priority_key or self.state_priority)
                states += staged_states[:missing]
                staged_states[:missing] = []
            else:
                l.debug("... staging stash disabled; unpickling states")
                states += self._unpickle(missing)

        if len(states) > self.max:
            l.debug("Too many states (%d/%d) in stash %s", len(states), self.max, self.src_stash)
            states.sort(key=self.priority_key or self.state_priority)
            staged_states += states[self.max :]
            states[self.max :] = []

        # if we have too few staged states, unpickle up to halfway between max and min
        if len(staged_states) < self.staging_min:
            l.debug("Too few states in staging stash (%s)", self.staging_stash)
            staged_states += self._unpickle((self.staging_min + self.staging_max) // 2 - len(staged_states))

        if len(staged_states) > self.staging_max:
            l.debug("Too many states in staging stash (%s)", self.staging_stash)
            self._pickle(staged_states[self.staging_max :])
            staged_states[self.staging_max :] = []

        simgr.stashes[self.src_stash] = states
        simgr.stashes[self.staging_stash] = staged_states
        return simgr

    @staticmethod
    def state_priority(state):
        return id(state)



# pylint:disable=arguments-differ,invalid-unary-operand-type
from __future__ import annotations
from typing import TYPE_CHECKING, cast
import logging

import angr.ailment as ailment
from angr.ailment.constant import UNDETERMINED_SIZE
from angr.sim_variable import SimVariable
import claripy
from unique_log_filter import UniqueLogFilter

from angr.engines.light.engine import SimEngineNostmtAIL
from angr.sim_type import SimTypeFunction
from angr.analyses.typehoon import typeconsts, typevars
from angr.analyses.typehoon.lifter import TypeLifter
from angr.utils.types import dereference_simtype_by_lib
from .engine_base import SimEngineVRBase, RichR

if TYPE_CHECKING:
    from .variable_recovery_fast import VariableRecoveryFastState  # noqa: F401


l = logging.getLogger(name=__name__)
l.addFilter(UniqueLogFilter())


class SimEngineVRAIL(
    SimEngineNostmtAIL["VariableRecoveryFastState", RichR[claripy.ast.BV | claripy.ast.FP], None, None],
    SimEngineVRBase["VariableRecoveryFastState", ailment.Block],
):
    """
    The engine for variable recovery on AIL.
    """

    def __init__(
        self,
        *args,
        type_lifter: TypeLifter,
        call_info=None,
        vvar_to_vvar: dict[int, int] | None,
        vvar_type_hints: dict[int, typeconsts.TypeConstant] | None = None,
        func_ret_var: SimVariable | None = None,
        **kwargs,
    ):
        super().__init__(*args, vvar_type_hints=vvar_type_hints, **kwargs)

        self._reference_spoffset: bool = False
        self.call_info = call_info or {}
        self.vvar_to_vvar = vvar_to_vvar
        self.type_lifter = type_lifter
        self.func_ret_var = func_ret_var

    def _mapped_vvarid(self, vvar_id: int) -> int | None:
        if self.vvar_to_vvar is not None and vvar_id in self.vvar_to_vvar:
            return self.vvar_to_vvar[vvar_id]
        return None

    def _process_block_end(self, block, stmt_data, whitelist):
        pass

    # Statement handlers

    def _handle_stmt_Assignment(self, stmt):
        dst_type = type(stmt.dst)

        if dst_type is ailment.Expr.Register:
            offset = stmt.dst.reg_offset
            data = self._expr(stmt.src)
            size = stmt.src.bits // 8

            if "write_size" in stmt.dst.tags and stmt.dst.tags["write_size"] > size:
                # zero-fill this register
                self._assign_to_register(
                    offset,
                    RichR(self.state.top(stmt.dst.tags["write_size"] * 8)),
                    stmt.dst.tags["write_size"],
                    create_variable=False,
                )

            self._assign_to_register(offset, data, size, src=stmt.src, dst=stmt.dst)

        elif dst_type is ailment.Expr.Tmp:
            # simply write to self.tmps
            data = self._expr(stmt.src)
            if data is None:
                return

            self.tmps[stmt.dst.tmp_idx] = data

        elif dst_type is ailment.Expr.VirtualVariable:
            data = self._expr(stmt.src)
            variable = self._assign_to_vvar(
                stmt.dst, data, src=stmt.src, dst=stmt.dst, vvar_id=self._mapped_vvarid(stmt.dst.varid)
            )

            if variable is not None and isinstance(stmt.src, ailment.Expr.Phi):
                # this is a phi node - we update variable manager's phi variable tracking
                for _, vvar in stmt.src.src_and_vvars:
                    if vvar is not None:
                        r = self._read_from_vvar(vvar, expr=stmt.src, vvar_id=self._mapped_vvarid(vvar.varid))
                        if r.variable is not None:
                            pv = self.state.variable_manager[self.func_addr]._phi_variables
                            if variable not in pv:
                                pv[variable] = set()
                            pv[variable].add(r.variable)

            if stmt.dst.was_stack and isinstance(stmt.dst.stack_offset, int):
                # store it to the stack region in case it's directly referenced later
                self._store(
                    RichR(self.state.stack_address(stmt.dst.stack_offset)),
                    data,
                    stmt.dst.bits // self.arch.byte_width,
                    atom=stmt.dst,
                )

        else:
            l.warning("Unsupported dst type %s.", dst_type)

    def _handle_stmt_WeakAssignment(self, stmt) -> None:
        src = self._expr(stmt.src)
        dst = self._expr(stmt.dst)
        if isinstance(src, RichR) and isinstance(dst, RichR) and src.typevar is not None and dst.typevar is not None:
            tc = typevars.Subtype(src.typevar, dst.typevar)
            self.state.add_type_constraint(tc)

    def _handle_stmt_CAS(self, stmt) -> None:
        self._expr(stmt.addr)
        self._expr(stmt.data_lo)
        if stmt.data_hi is not None:
            self._expr(stmt.data_hi)
        self._expr(stmt.expd_lo)
        if stmt.expd_hi is not None:
            self._expr(stmt.expd_hi)

    def _handle_stmt_Store(self, stmt: ailment.Stmt.Store):
        addr_r = self._expr_bv(stmt.addr)
        data = self._expr(stmt.data)
        size = stmt.size
        self._store(addr_r, data, size, atom=stmt)

    def _handle_stmt_Jump(self, stmt: ailment.Stmt.Jump):
        if not isinstance(stmt.target, ailment.Expr.Const):
            self._expr(stmt.target)

    def _handle_stmt_ConditionalJump(self, stmt):
        self._expr(stmt.condition)

    def _handle_expr_Tmp(self, expr):
        try:
            return self.tmps[expr.tmp_idx]
        except KeyError:
            return self._top(expr.bits)

    def _handle_expr_MultiStatementExpression(self, expr):
        for stmt in expr.statements:
            self._stmt(stmt)
        return self._expr(expr.expr)

    def _handle_expr_Call(self, expr):
        target = expr.target
        args = []
        if expr.args:
            for arg in expr.args:
                self._reference_spoffset = True
                richr = self._expr(arg)
                self._reference_spoffset = False
                args.append(richr)

        ret_expr_bits = expr.bits

        if isinstance(target, ailment.Expr.Expression) and not isinstance(
            target, (ailment.Expr.Const, ailment.Expr.DirtyExpression)
        ):
            # this is a dynamically calculated call target
            target_expr = self._expr(target)
            funcaddr_typevar = target_expr.typevar
            if isinstance(funcaddr_typevar, typevars.TypeVariable):
                load_typevar = self._create_access_typevar(funcaddr_typevar, False, self.arch.bytes, 0)
                self.state.add_type_constraint(typevars.Subtype(funcaddr_typevar, load_typevar))
        elif isinstance(target, str):
            # special handling for some intrinsics
            match target:
                case (
                    "InterlockedExchange8"
                    | "InterlockedExchange16"
                    | "InterlockedExchange"
                    | "InterlockedExchange64"
                    | "InterlockedCompareExchange16"
                    | "InterlockedCompareExchange"
                    | "InterlockedCompareExchange64"
                    | "InterlockedCompareExchange128"
                    | "InterlockedExchangeAdd"
                    | "InterlockedExchangeAdd64"
                ):
                    arg_tv = (
                        args[0].typevar
                        if args[0].typevar is not None
                        else args[1].typevar if args[1].typevar is not None else None
                    )
                    if arg_tv is not None:
                        ret_ty = self._create_access_typevar(
                            arg_tv, False, args[0].data.size() // self.arch.byte_width, 0
                        )
                        return RichR(self.state.top(ret_expr_bits), typevar=ret_ty)
                case (
                    "InterlockedIncrement"
                    | "InterlockedIncrement16"
                    | "InterlockedIncrement64"
                    | "InterlockedDecrement"
                    | "InterlockedDecrement16"
                    | "InterlockedDecrement64"
                ):
                    arg_tv = args[0].typevar if args[0].typevar is not None else None
                    if arg_tv is not None:
                        ret_ty = self._create_access_typevar(
                            arg_tv, False, args[0].data.size() // self.arch.byte_width, 0
                        )
                        return RichR(self.state.top(ret_expr_bits), typevar=ret_ty)

        # discover the prototype
        prototype: SimTypeFunction | None = None
        prototype_libname: str | None = None
        if expr.prototype is not None:
            prototype = expr.prototype
        if isinstance(expr.target, ailment.Expr.Const):
            func_addr = expr.target.value
            if isinstance(func_addr, self.kb.functions.address_types) and func_addr in self.kb.functions:
                func = self.kb.functions[func_addr]
                if prototype is None:
                    prototype = func.prototype
                prototype_libname = func.prototype_libname

        ret_ty = None
        if prototype is not None:
            # add type constraints
            if args:
                for arg, arg_type in zip(args, prototype.args):
                    if arg.typevar is not None:
                        arg_type = (
                            dereference_simtype_by_lib(arg_type, prototype_libname) if prototype_libname else arg_type
                        )
                        arg_ty = self.type_lifter.lift(arg_type)
                        type_constraint = typevars.Subtype(arg.typevar, arg_ty)
                        self.state.add_type_constraint(type_constraint)
            if not expr.tags.get("is_prototype_guessed", True):
                return_ty = self.type_lifter.lift(prototype.returnty)  # type:ignore
                ret_ty = typevars.TypeVariable()
                if not isinstance(ret_ty, typeconsts.BottomType):
                    type_constraint = typevars.Subtype(ret_ty, return_ty)
                    self.state.add_type_constraint(type_constraint)

        if ret_ty is None:
            ret_ty = typevars.TypeVariable()

        return RichR(self.state.top(ret_expr_bits), typevar=ret_ty)

    def _handle_stmt_Call(self, stmt):
        target = stmt.target
        args: list[RichR] = []
        if stmt.args:
            for arg in stmt.args:
                self._reference_spoffset = True
                richr = self._expr(arg)
                self._reference_spoffset = False
                args.append(richr)

        ret_expr_bits = self.state.arch.bits
        create_variable = True

        # this is a call statement. we need to update the return value register later
        ret_expr = stmt.ret_expr
        if ret_expr is not None:
            if ret_expr.category == ailment.Expr.VirtualVariableCategory.REGISTER:
                ret_expr_bits = ret_expr.bits
        else:
            # the return expression is not used, so we treat this call as not returning anything
            create_variable = False

        if isinstance(target, ailment.Expr.Expression) and not isinstance(
            target, (ailment.Expr.Const, ailment.Expr.DirtyExpression)
        ):
            # this is a dynamically calculated call target
            target_expr = self._expr(target)
            funcaddr_typevar = target_expr.typevar
            if isinstance(funcaddr_typevar, typevars.TypeVariable):
                load_typevar = self._create_access_typevar(funcaddr_typevar, False, self.arch.bytes, 0)
                self.state.add_type_constraint(typevars.Subtype(funcaddr_typevar, load_typevar))

        # discover the prototype
        prototype: SimTypeFunction | None = None
        prototype_libname: str | None = None
        if stmt.prototype is not None:
            prototype = stmt.prototype
        if isinstance(stmt.target, ailment.Expr.Const):
            func_addr = stmt.target.value
            if isinstance(func_addr, self.kb.functions.address_types) and func_addr in self.kb.functions:
                func = self.kb.functions[func_addr]
                if prototype is None:
                    prototype = func.prototype
                prototype_libname = func.prototype_libname

        ret_ty = None
        if prototype is not None:
            # add type constraints
            if args:
                for arg, arg_type in zip(args, prototype.args):
                    if arg.typevar is not None:
                        arg_type = (
                            dereference_simtype_by_lib(arg_type, prototype_libname) if prototype_libname else arg_type
                        )
                        arg_ty = self.type_lifter.lift(arg_type)
                        if arg.typevar is not None and isinstance(
                            arg_ty, (typeconsts.TypeConstant, typevars.TypeVariable, typevars.DerivedTypeVariable)
                        ):
                            type_constraint = typevars.Subtype(arg.typevar, arg_ty)
                            self.state.add_type_constraint(type_constraint)
            return_ty = self.type_lifter.lift(prototype.returnty)  # type:ignore
            ret_ty = typevars.TypeVariable()
            if not isinstance(ret_ty, typeconsts.BottomType):
                type_constraint = typevars.Subtype(return_ty, ret_ty)
                self.state.add_type_constraint(type_constraint)

        if ret_ty is None:
            ret_ty = typevars.TypeVariable()

        # TODO: Expose it as an option
        return_value_use_full_width_reg = True

        # update the return value register
        if isinstance(ret_expr, ailment.Expr.VirtualVariable):
            expr_bits = ret_expr_bits
            self._assign_to_vvar(
                ret_expr,
                RichR(self.state.top(expr_bits), typevar=ret_ty),
                dst=ret_expr,
                create_variable=create_variable,
                vvar_id=self._mapped_vvarid(ret_expr.varid),
            )
        elif isinstance(ret_expr, ailment.Expr.Register):
            l.warning("Left-over register found in call.ret_expr.")
            expr_bits = self.state.arch.bits if return_value_use_full_width_reg else ret_expr_bits
            self._assign_to_register(
                ret_expr.reg_offset,
                RichR(self.state.top(expr_bits), typevar=ret_ty),
                expr_bits // self.arch.byte_width,
                dst=ret_expr,
                create_variable=create_variable,
            )

    def _handle_stmt_Return(self, stmt):
        if stmt.ret_exprs:
            if self.func_ret_var is None:
                ret_typevar = None
            elif self.state.typevars.has_type_variable_for(self.func_ret_var):
                ret_typevar = self.state.typevars.get_type_variable(self.func_ret_var)
            else:
                ret_typevar = typevars.TypeVariable()
                self.state.typevars.add_type_variable(self.func_ret_var, ret_typevar)

            for ret_expr in stmt.ret_exprs:
                src = self._expr(ret_expr)
                if isinstance(src, RichR) and src.typevar is not None and ret_typevar is not None:
                    if src.type_constraints is not None:
                        for tc in src.type_constraints:
                            self.state.add_type_constraint(tc)
                    tc = typevars.Subtype(ret_typevar, src.typevar)
                    self.state.add_type_constraint(tc)

    def _handle_expr_DirtyExpression(self, expr: ailment.Expr.DirtyExpression) -> RichR:
        for op in expr.operands:
            self._expr(op)
        if expr.guard:
            self._expr(expr.guard)
        if expr.maddr:
            self._expr(expr.maddr)
        return RichR(self.state.top(expr.bits))

    def _handle_expr_VEXCCallExpression(self, expr: ailment.Expr.VEXCCallExpression) -> RichR:
        for op in expr.operands:
            self._expr(op)
        return RichR(self.state.top(expr.bits))

    # Expression handlers

    def _expr_bv(self, expr: ailment.expression.Expression) -> RichR[claripy.ast.BV]:
        result = self._expr(expr)
        assert isinstance(result.data, claripy.ast.BV)
        return cast(RichR[claripy.ast.BV], result)

    def _expr_fp(self, expr: ailment.expression.Expression) -> RichR[claripy.ast.FP]:
        result = self._expr(expr)
        assert isinstance(result.data, claripy.ast.FP)
        return cast(RichR[claripy.ast.FP], result)

    def _expr_pair(
        self, expr1: ailment.expression.Expression, expr2: ailment.expression.Expression
    ) -> tuple[RichR[claripy.ast.BV], RichR[claripy.ast.BV]] | tuple[RichR[claripy.ast.FP], RichR[claripy.ast.FP]]:
        result1 = self._expr(expr1)
        result2 = self._expr(expr2)
        assert type(result1.data) is type(result2.data)
        return result1, result2  # type: ignore

    def _handle_expr_Register(self, expr):
        offset = expr.reg_offset
        size = expr.bits // 8

        return self._read_from_register(offset, size, expr=expr)

    def _handle_expr_Load(self, expr):
        addr_r = self._expr_bv(expr.addr)
        size = expr.size

        if size != UNDETERMINED_SIZE:
            return self._load(addr_r, size, expr=expr)
        return self._top(8)

    def _handle_expr_VirtualVariable(self, expr: ailment.Expr.VirtualVariable):
        return self._read_from_vvar(expr, expr=expr, vvar_id=self._mapped_vvarid(expr.varid))

    def _handle_expr_Phi(self, expr: ailment.Expr.Phi):
        tvs = set()
        for _, vvar in expr.src_and_vvars:
            if vvar is not None:
                r = self._read_from_vvar(vvar, expr=vvar, vvar_id=self._mapped_vvarid(vvar.varid))
                if r.typevar is not None:
                    tvs.add(r.typevar)

        tv = typevars.TypeVariable()
        for tv_ in tvs:
            self.state.add_type_constraint(typevars.Subtype(tv, tv_))
        return RichR(self.state.top(expr.bits), typevar=tv)

    def _handle_expr_Const(self, expr: ailment.Expr.Const):
        if isinstance(expr.value, float):
            v = claripy.FPV(expr.value, claripy.FSORT_DOUBLE if expr.bits == 64 else claripy.FSORT_FLOAT).to_bv()
            ty = typeconsts.float_type(expr.bits)
        else:
            if self.project.loader.find_segment_containing(expr.value) is not None:
                r = self._load_from_global(expr.value, 1, expr=expr)
                ty = r.typevar
            elif expr.value == 0 and expr.bits == self.arch.bits:
                # this can be viewed as a NULL
                ty = (
                    typeconsts.Pointer64(typeconsts.TopType())
                    if self.arch.bits == 64
                    else typeconsts.Pointer32(typeconsts.TopType())
                )
            else:
                ty = typeconsts.int_type(expr.bits)
            v = claripy.BVV(expr.value, expr.bits)
        r = RichR(v, typevar=ty)
        codeloc = self._codeloc()
        self._ensure_variable_existence(r, codeloc)
        self._reference(r, codeloc)
        return r

    def _handle_expr_Convert(self, expr: ailment.Expr.Convert):
        r = self._expr(expr.operand)
        typevar = None
        if r.typevar is not None:
            if isinstance(r.typevar, typevars.DerivedTypeVariable) and isinstance(
                r.typevar.one_label, typevars.ConvertTo
            ):
                # there is already a conversion - overwrite it
                if not isinstance(r.typevar.type_var, typeconsts.TypeConstant):
                    typevar = typevars.new_dtv(r.typevar.type_var, label=typevars.ConvertTo(expr.to_bits))
            else:
                if not isinstance(r.typevar, typeconsts.TypeConstant):
                    typevar = typevars.new_dtv(r.typevar, label=typevars.ConvertTo(expr.to_bits))

        return RichR(self.state.top(expr.to_bits), typevar=typevar)

    def _handle_expr_Reinterpret(self, expr: ailment.Expr.Reinterpret):
        r = self._expr(expr.operand)
        typevar = None
        if r.typevar is not None:
            if isinstance(r.typevar, typevars.DerivedTypeVariable) and isinstance(
                r.typevar.one_label, typevars.ReinterpretAs
            ):
                # there is already a reinterpretas - overwrite it
                typevar = typevars.new_dtv(r.typevar.type_var, label=typevars.ReinterpretAs(expr.to_type, expr.to_bits))
            elif isinstance(r.typevar, typevars.TypeVariable):
                typevar = typevars.new_dtv(r.typevar, label=typevars.ReinterpretAs(expr.to_type, expr.to_bits))

        return RichR(self.state.top(expr.to_bits), typevar=typevar)

    def _handle_expr_StackBaseOffset(self, expr: ailment.Expr.StackBaseOffset):
        refbase_typevar = self.state.stack_offset_typevars.get(expr.offset, None)
        if refbase_typevar is None:
            # allocate a new type variable
            refbase_typevar = typevars.TypeVariable()
            self.state.stack_offset_typevars[expr.offset] = refbase_typevar

        ref_typevar = typevars.TypeVariable()
        access_derived_typevar = self._create_access_typevar(ref_typevar, False, None, 0)
        load_constraint = typevars.Subtype(refbase_typevar, access_derived_typevar)
        self.state.add_type_constraint(load_constraint)

        value_v = self.state.stack_address(expr.offset)
        richr = RichR(value_v, typevar=ref_typevar)
        codeloc = self._codeloc()
        self._ensure_variable_existence(richr, codeloc, src_expr=expr)
        if self._reference_spoffset:
            self._reference(richr, codeloc, src=expr)
        return richr

    def _handle_unop_Reference(self, expr: ailment.Expr.UnaryOp):
        if isinstance(expr.operand, ailment.Expr.VirtualVariable) and expr.operand.was_stack:

            refbase_typevar = None
            off = expr.operand.stack_offset

            # does this variable exist?
            value: claripy.ast.BV | None = self.vvar_region.get(expr.operand.varid, None)
            if value is not None:
                for _, var in self.state.extract_variables(value):
                    if var in self.state.typevars:
                        refbase_typevar = self.state.typevars[var]
                        break

            if refbase_typevar is None:
                if off in self.state.stack_offset_typevars:
                    refbase_typevar = self.state.stack_offset_typevars[off]
                else:
                    # allocate a new type variable
                    refbase_typevar = typevars.TypeVariable()
                    self.state.stack_offset_typevars[off] = refbase_typevar

            ref_typevar = typevars.TypeVariable()
            access_derived_typevar = self._create_access_typevar(ref_typevar, False, None, 0)
            load_constraint = typevars.Subtype(refbase_typevar, access_derived_typevar)
            self.state.add_type_constraint(load_constraint)

            value_v = self.state.stack_address(off)
            richr = RichR(value_v, typevar=ref_typevar)
            codeloc = self._codeloc()
            self._ensure_variable_existence(richr, codeloc, src_expr=expr.operand)
            if self._reference_spoffset:
                self._reference(richr, codeloc, src=expr.operand)
            return richr
        return RichR(self.state.top(expr.bits))

    def _handle_expr_BasePointerOffset(self, expr):
        # TODO
        return self._top(expr.bits)

    def _handle_expr_ITE(self, expr: ailment.Expr.ITE):
        self._expr(expr.cond)  # cond
        r0 = self._expr(expr.iftrue)
        r1 = self._expr(expr.iffalse)

        type_constraints = set()
        tv = typevars.TypeVariable()
        if r0.typevar is not None:
            type_constraints.add(typevars.Subtype(tv, r0.typevar))
        if r1.typevar is not None:
            type_constraints.add(typevars.Subtype(tv, r1.typevar))

        return RichR(self.state.top(expr.bits), typevar=tv, type_constraints=type_constraints)

    def _handle_binop_Add(self, expr):
        arg0, arg1 = expr.operands
        r0, r1 = self._expr_pair(arg0, arg1)
        compute = r0.data + r1.data if r0.data.size() == r1.data.size() else self.state.top(expr.bits)  # type: ignore

        type_constraints = set()
        # create a new type variable and add constraints accordingly
        r0_typevar = r0.typevar if r0.typevar is not None else typevars.TypeVariable()

        typevar = None
        if r1.data.concrete:
            # addition with constants. create a derived type variable
            if isinstance(r0_typevar, typevars.TypeVariable):
                typevar = typevars.new_dtv(r0_typevar, label=typevars.AddN(r1.data.concrete_value))
        elif r1.typevar is not None:
            typevar = typevars.TypeVariable()
            type_constraints.add(typevars.Add(r0_typevar, r1.typevar, typevar))
        else:
            typevar = None

        return RichR(compute, typevar=typevar, type_constraints=type_constraints)

    def _handle_binop_Sub(self, expr):
        arg0, arg1 = expr.operands
        r0, r1 = self._expr_pair(arg0, arg1)
        compute = r0.data - r1.data  # type: ignore

        type_constraints = set()
        typevar = None
        if r0.typevar is not None and r1.data.concrete and isinstance(r0.typevar, typevars.TypeVariable):
            typevar = typevars.new_dtv(r0.typevar, label=typevars.SubN(r1.data.concrete_value))
        else:
            typevar = typevars.TypeVariable()
            if r0.typevar is not None and r1.typevar is not None:
                type_constraints.add(typevars.Sub(r0.typevar, r1.typevar, typevar))

        return RichR(
            compute,
            typevar=typevar,
            type_constraints=type_constraints,
        )

    def _handle_binop_Mul(self, expr):
        arg0, arg1 = expr.operands
        r0, r1 = self._expr_pair(arg0, arg1)

        result_size = arg0.bits
        if r0.data.concrete or r1.data.concrete:
            # constants
            result_size = arg0.bits
            compute = r0.data * r1.data  # type: ignore
            return RichR(compute, typevar=typeconsts.int_type(result_size), type_constraints=None)

        r = self.state.top(expr.bits)
        return RichR(
            r,
            typevar=r0.typevar,
        )

    def _handle_binop_Mull(self, expr):
        arg0, arg1 = expr.operands

        r0 = self._expr_bv(arg0)
        r1 = self._expr_bv(arg1)

        if r0.data.concrete and r1.data.concrete:
            # constants
            result_size = expr.bits
            if r0.data.size() < result_size:
                if expr.signed:
                    r0.data = claripy.SignExt(result_size - r0.data.size(), r0.data)
                else:
                    r0.data = claripy.ZeroExt(result_size - r0.data.size(), r0.data)
            if r1.data.size() < result_size:
                if expr.signed:
                    r1.data = claripy.SignExt(result_size - r1.data.size(), r1.data)
                else:
                    r1.data = claripy.ZeroExt(result_size - r1.data.size(), r1.data)
            return RichR(r0.data * r1.data, typevar=typeconsts.int_type(result_size), type_constraints=None)

        r = self.state.top(expr.bits)
        return RichR(
            r,
            typevar=r0.typevar,  # FIXME: the size is probably changed
        )

    def _handle_binop_Div(self, expr):
        arg0, arg1 = expr.operands

        r0 = self._expr_bv(arg0)
        r1 = self._expr_bv(arg1)
        from_size = expr.bits
        to_size = r1.bits

        if expr.floating_point:
            quotient = self.state.top(to_size)
        else:
            if (r1.data == 0).is_true():
                quotient = self.state.top(to_size)
            elif expr.signed:
                quotient = claripy.SDiv(r0.data, claripy.SignExt(from_size - to_size, r1.data))
            else:
                quotient = r0.data // claripy.ZeroExt(from_size - to_size, r1.data)

        return RichR(
            quotient,
            # | typevar=r0.typevar,  # FIXME: Handle typevars for Div
        )

    def _handle_binop_Mod(self, expr):
        arg0, arg1 = expr.operands

        r0 = self._expr_bv(arg0)
        r1 = self._expr_bv(arg1)
        result_size = expr.bits

        if expr.floating_point:
            remainder = self.state.top(result_size)
        else:
            if (r1.data == 0).is_true():
                remainder = self.state.top(result_size)
            elif expr.signed:
                remainder = r0.data.SMod(r1.data)
            else:
                remainder = r0.data % r1.data

        # truncation if necessary
        if remainder.size() > result_size:
            remainder = claripy.Extract(result_size - 1, 0, remainder)

        return RichR(
            remainder,
            # | typevar=r0.typevar,  # FIXME: Handle typevars for Mod
        )

    def _handle_binop_Xor(self, expr):
        arg0, arg1 = expr.operands

        r0 = self._expr_bv(arg0)
        r1 = self._expr_bv(arg1)

        if r0.data.concrete and r1.data.concrete:
            # constants
            result_size = arg0.bits
            return RichR(r0.data ^ r1.data, typevar=typeconsts.int_type(result_size), type_constraints=None)

        # xor does not transfer type variables; instead, it forces both operands to be unsigned integers
        if isinstance(r0.typevar, typevars.TypeVariable):
            tc = typevars.Subtype(r0.typevar, typeconsts.int_type(r0.data.size()))
            self.state.add_type_constraint(tc)
        if isinstance(r1.typevar, typevars.TypeVariable):
            tc = typevars.Subtype(r1.typevar, typeconsts.int_type(r1.data.size()))
            self.state.add_type_constraint(tc)

        r = self.state.top(expr.bits)
        return RichR(r)

    def _handle_binop_Shl(self, expr):
        arg0, arg1 = expr.operands

        r0 = self._expr_bv(arg0)
        r1 = self._expr_bv(arg1)
        result_size = arg0.bits

        if not r1.data.concrete:
            # we don't support symbolic shiftamount
            r = self.state.top(result_size)
            return RichR(r)

        shiftamount = r1.data.concrete_value
        return RichR(r0.data << shiftamount, typevar=typeconsts.int_type(result_size), type_constraints=None)

    def _handle_binop_Shr(self, expr):
        arg0, arg1 = expr.operands

        r0 = self._expr_bv(arg0)
        r1 = self._expr_bv(arg1)
        result_size = arg0.bits

        if not r1.data.concrete:
            # we don't support symbolic shiftamount
            r = self.state.top(result_size)
            return RichR(r)

        shiftamount = r1.data.concrete_value

        return RichR(
            claripy.LShR(r0.data, shiftamount), typevar=typeconsts.int_type(result_size), type_constraints=None
        )

    def _handle_binop_Sal(self, expr):
        arg0, arg1 = expr.operands

        r0 = self._expr_bv(arg0)
        r1 = self._expr_bv(arg1)
        result_size = arg0.bits

        if not r1.data.concrete:
            # we don't support symbolic shiftamount
            r = self.state.top(result_size)
            return RichR(r)

        shiftamount = r1.data.concrete_value

        return RichR(r0.data << shiftamount, typevar=typeconsts.int_type(result_size), type_constraints=None)

    def _handle_binop_Sar(self, expr):
        arg0, arg1 = expr.operands

        r0 = self._expr_bv(arg0)
        r1 = self._expr_bv(arg1)
        result_size = arg0.bits

        if not r1.data.concrete:
            # we don't support symbolic shiftamount
            r = self.state.top(result_size)
            return RichR(r)

        shiftamount = r1.data.concrete_value

        return RichR(r0.data >> shiftamount, typevar=typeconsts.int_type(result_size), type_constraints=None)

    def _handle_binop_And(self, expr):
        arg0, arg1 = expr.operands

        r0 = self._expr_bv(arg0)
        r1 = self._expr_bv(arg1)

        result_size = arg0.bits
        if r0.data.concrete and r1.data.concrete:
            return RichR(
                r0.data & r1.data,
                typevar=typeconsts.int_type(result_size),
                type_constraints=None,
            )

        r = self.state.top(expr.bits)
        return RichR(r, typevar=typeconsts.int_type(result_size))

    def _handle_binop_Or(self, expr):
        arg0, arg1 = expr.operands

        r0 = self._expr_bv(arg0)
        r1 = self._expr_bv(arg1)

        result_size = arg0.bits
        if r0.data.concrete and r1.data.concrete:
            return RichR(
                r0.data | r1.data,
                typevar=typeconsts.int_type(result_size),
                type_constraints=None,
            )

        r = self.state.top(expr.bits)
        return RichR(r, typevar=typeconsts.int_type(result_size))

    def _handle_binop_LogicalAnd(self, expr):
        arg0, arg1 = expr.operands

        r0 = self._expr_bv(arg0)
        _ = self._expr_bv(arg1)
        r = self.state.top(expr.bits)
        return RichR(r, typevar=r0.typevar)

    def _handle_binop_LogicalOr(self, expr):
        arg0, arg1 = expr.operands

        r0 = self._expr_bv(arg0)
        _ = self._expr_bv(arg1)
        r = self.state.top(expr.bits)
        return RichR(r, typevar=r0.typevar)

    def _handle_binop_LogicalXor(self, expr):
        arg0, arg1 = expr.operands

        r0 = self._expr_bv(arg0)
        _ = self._expr_bv(arg1)
        r = self.state.top(expr.bits)
        return RichR(r, typevar=r0.typevar)

    def _handle_binop_Rol(self, expr):
        arg0, arg1 = expr.operands

        _ = self._expr_bv(arg0)
        _ = self._expr_bv(arg1)
        result_size = arg0.bits

        r = self.state.top(result_size)
        return RichR(r)

    def _handle_binop_Ror(self, expr):
        arg0, arg1 = expr.operands

        _ = self._expr_bv(arg0)
        _ = self._expr_bv(arg1)
        result_size = arg0.bits

        r = self.state.top(result_size)
        return RichR(r)

    def _handle_binop_Concat(self, expr):
        arg0, arg1 = expr.operands

        _ = self._expr_bv(arg0)
        _ = self._expr_bv(arg1)

        # TODO: Model the operation. Don't lose type constraints
        return RichR(self.state.top(expr.bits))

    def _handle_binop_Default(self, expr):
        arg0, arg1 = expr.operands

        self._expr(arg0)
        self._expr(arg1)

        return RichR(self.state.top(expr.bits))

    _handle_binop_AddF = _handle_binop_Default
    _handle_binop_SubF = _handle_binop_Default
    _handle_binop_SubV = _handle_binop_Default
    _handle_binop_MulF = _handle_binop_Default
    _handle_binop_DivF = _handle_binop_Default
    _handle_binop_DivV = _handle_binop_Default
    _handle_binop_AddV = _handle_binop_Default
    _handle_binop_MulV = _handle_binop_Default
    _handle_binop_MulHiV = _handle_binop_Default
    _handle_binop_Carry = _handle_binop_Default
    _handle_binop_Borrow = _handle_binop_Default
    _handle_binop_SCarry = _handle_binop_Default
    _handle_binop_SBorrow = _handle_binop_Default
    _handle_binop_InterleaveLOV = _handle_binop_Default
    _handle_binop_InterleaveHIV = _handle_binop_Default
    _handle_binop_CasCmpEQ = _handle_binop_Default
    _handle_binop_CasCmpNE = _handle_binop_Default
    _handle_binop_ExpCmpNE = _handle_binop_Default
    _handle_binop_SarNV = _handle_binop_Default
    _handle_binop_ShrNV = _handle_binop_Default
    _handle_binop_ShlNV = _handle_binop_Default
    _handle_binop_PermV = _handle_binop_Default
    _handle_binop_Set = _handle_binop_Default
    _handle_binop_MaxV = _handle_binop_Default
    _handle_binop_MinV = _handle_binop_Default
    _handle_binop_QAddV = _handle_binop_Default
    _handle_binop_QNarrowBinV = _handle_binop_Default
    _handle_binop_CmpEQ = _handle_binop_Default
    _handle_binop_CmpNE = _handle_binop_Default
    _handle_binop_CmpLT = _handle_binop_Default
    _handle_binop_CmpLE = _handle_binop_Default
    _handle_binop_CmpGT = _handle_binop_Default
    _handle_binop_CmpGE = _handle_binop_Default
    _handle_binop_CmpORD = _handle_binop_Default
    _handle_binop_CmpEQV = _handle_binop_Default
    _handle_binop_CmpNEV = _handle_binop_Default
    _handle_binop_CmpGEV = _handle_binop_Default
    _handle_binop_CmpGTV = _handle_binop_Default
    _handle_binop_CmpLEV = _handle_binop_Default
    _handle_binop_CmpLTV = _handle_binop_Default
    _handle_binop_CmpF = _handle_binop_Default

    def _handle_unop_Not(self, expr):
        arg = expr.operands[0]
        expr = self._expr_bv(arg)

        result_size = arg.bits

        if expr.data.concrete:
            return RichR(
                ~expr.data,
                typevar=typeconsts.int_type(result_size),
                type_constraints=None,
            )

        r = self.state.top(result_size)
        return RichR(r, typevar=expr.typevar)

    def _handle_unop_Neg(self, expr):
        arg = expr.operands[0]
        expr = self._expr_bv(arg)

        result_size = arg.bits

        if expr.data.concrete:
            return RichR(
                -expr.data,
                typevar=typeconsts.int_type(result_size),
                type_constraints=None,
            )

        r = self.state.top(result_size)
        return RichR(r, typevar=expr.typevar)

    def _handle_unop_BitwiseNeg(self, expr):
        arg = expr.operands[0]
        expr = self._expr_bv(arg)

        result_size = arg.bits

        if expr.data.concrete:
            return RichR(
                ~expr.data,
                typevar=typeconsts.int_type(result_size),
                type_constraints=None,
            )

        r = self.state.top(result_size)
        return RichR(r, typevar=expr.typevar)

    def _handle_unop_Default(self, expr):
        self._expr(expr.operands[0])
        return RichR(self.state.top(expr.bits))

    _handle_unop_Dereference = _handle_unop_Default
    _handle_unop_Clz = _handle_unop_Default
    _handle_unop_Ctz = _handle_unop_Default
    _handle_unop_GetMSBs = _handle_unop_Default
    _handle_unop_unpack = _handle_unop_Default
    _handle_unop_Sqrt = _handle_unop_Default
    _handle_unop_RSqrtEst = _handle_unop_Default



from __future__ import annotations
import angr


class gethostbyname(angr.SimProcedure):
    def run(self, name):
        malloc = angr.SIM_PROCEDURES["libc"]["malloc"]
        int_size_bits = self.arch.sizeof["int"]
        int_size = int_size_bits // 8
        ptr_size_bits = self.arch.bits
        ptr_size = ptr_size_bits // 8
        hostent_size = (2 * int_size) + (3 * ptr_size)
        place = self.inline_call(malloc, hostent_size).ret_expr
        self.state.memory.store(
            place,
            self.state.solver.BVS("h_name", ptr_size_bits, key=("api", "gethostbyname", "h_name")),
            endness="Iend_LE",
        )
        next_addr = place + ptr_size
        self.state.memory.store(
            next_addr,
            self.state.solver.BVS("h_aliases", ptr_size_bits, key=("api", "gethostbyname", "h_aliases")),
            endness="Iend_LE",
        )
        next_addr += ptr_size
        self.state.memory.store(
            next_addr,
            self.state.solver.BVS("h_addrtype", int_size_bits, key=("api", "gethostbyname", "h_addrtype")),
            endness="Iend_LE",
        )
        next_addr += int_size
        self.state.memory.store(
            next_addr,
            self.state.solver.BVS("h_length", int_size_bits, key=("api", "gethostbyname", "h_length")),
            endness="Iend_LE",
        )
        next_addr += int_size
        self.state.memory.store(
            next_addr,
            self.state.solver.BVS("h_addr_list", ptr_size_bits, key=("api", "gethostbyname", "h_addr_list")),
            endness="Iend_LE",
        )
        return place



from __future__ import annotations

from collections.abc import Iterable, Callable
from typing import Any

from angr.storage.memory_mixins.memory_mixin import MemoryMixin


class TopMergerMixin(MemoryMixin):
    """
    A memory mixin for merging values in memory to TOP.
    """

    def __init__(self, *args, top_func=None, **kwargs):
        self._top_func: Callable = top_func

        super().__init__(*args, **kwargs)

    def _merge_values(self, values: Iterable[tuple[Any, Any]], merged_size: int, **kwargs):
        return self._top_func(merged_size * self.state.arch.byte_width)

    def copy(self, memo=None):
        copied = super().copy(memo)
        copied._top_func = self._top_func
        return copied



from __future__ import annotations
from typing import TYPE_CHECKING
import math
import logging

from angr.analyses.analysis import Analysis, AnalysesHub
from angr.knowledge_plugins.cfg import CFGModel


if TYPE_CHECKING:
    from cle import Section

_l = logging.getLogger(__name__)


class PackingDetector(Analysis):
    """
    This analysis detects if a binary is likely packed or not. We may extend it to identify which packer is in use in
    the future.
    """

    PACKED_MIN_BYTES = 256
    PACKED_ENTROPY_MIN_THRESHOLD = 0.88

    def __init__(self, cfg: CFGModel | None = None, region_size_threshold: int = 0x20):
        self.packed: bool = False
        self.region_size_threshold: int = region_size_threshold

        if cfg is None:
            _l.warning(
                "PackingDetector is using a most accurate CFG model in the knowledge base. We assume it is "
                "generated with force_smart_scan=False and force_complete_scan=False."
            )
            self._cfg = self.kb.cfgs.get_most_accurate()
        else:
            self._cfg = cfg

        self.analyze()

    def analyze(self):
        # assume we already have a CFG with complete scanning disabled
        # collect all regions that are not covered by the CFG in r+x sections, and then compute the entropy. we believe
        # the binary is packed if it is beyond a threshold

        covered_regions: list[tuple[int, int]] = []
        last_known_section: Section | None = None
        for node in sorted(self._cfg.nodes(), key=lambda n: n.addr):
            section = None
            if last_known_section is not None and last_known_section.contains_addr(node.addr):
                section = last_known_section
            if section is None:
                section = self.project.loader.find_section_containing(node.addr)
                if section is None:
                    # this node does not belong to any known section - ignore it
                    continue
                if section.is_readable and section.is_executable:
                    last_known_section = section

            if section is None:
                # the node does not belong to any section. ignore it
                continue

            if node.size == 0:
                # ignore empty nodes
                continue

            if not covered_regions:
                covered_regions.append((node.addr, node.addr + node.size))
            else:
                last_item = covered_regions[-1]
                if last_item[0] <= node.addr <= last_item[1] < node.addr + node.size:
                    # update the last item
                    covered_regions[-1] = last_item[0], node.addr + node.size
                else:
                    # add a new item
                    covered_regions.append((node.addr, node.addr + node.size))

        # now we get the uncovered regions
        uncovered_regions: list[tuple[int, int]] = self._get_uncovered_regions(covered_regions)

        # compute entropy
        total_bytes, entropy = self._compute_entropy(uncovered_regions)

        self.packed = total_bytes >= self.PACKED_MIN_BYTES and entropy >= self.PACKED_ENTROPY_MIN_THRESHOLD

    def _get_uncovered_regions(self, covered_regions: list[tuple[int, int]]) -> list[tuple[int, int]]:
        # FIXME: We only support binaries with sections. Add support for segments in the future
        all_executable_sections = [
            sec
            for sec in self.project.loader.main_object.sections
            if sec.is_executable and sec.is_readable and not sec.only_contains_uninitialized_data
        ]
        all_executable_sections = sorted(all_executable_sections, key=lambda sec: sec.vaddr)
        idx = 0

        uncovered_regions: list[tuple[int, int]] = []
        for section in all_executable_sections:
            if idx >= len(covered_regions):
                if section.memsize > self.region_size_threshold:
                    uncovered_regions.append((section.vaddr, section.vaddr + section.memsize))
            else:
                i = idx
                last_end = section.vaddr
                while i < len(covered_regions):
                    region_start, region_end = covered_regions[i]
                    if region_end >= section.vaddr + section.memsize:
                        # move on to the next section
                        break
                    if last_end < region_start and region_start - last_end > self.region_size_threshold:
                        uncovered_regions.append((last_end, region_start))
                    i += 1
                    last_end = max(last_end, region_end)
                idx = i

        return uncovered_regions

    def _compute_entropy(self, regions: list[tuple[int, int]]) -> tuple[int, float]:
        byte_counts = [0] * 256

        for start, end in regions:
            for b in self.project.loader.memory.load(start, end - start):
                byte_counts[b] += 1

        total = sum(byte_counts)
        if total == 0:
            return 0, 0.0

        entropy = 0.0
        for count in byte_counts:
            if count == 0:
                continue
            p = 1.0 * count / total
            entropy -= p * math.log(p, 256)

        return total, entropy


AnalysesHub.register_default("PackingDetector", PackingDetector)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: angr/protos/variables.proto
# Protobuf Python Version: 6.33.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    0,
    '',
    'angr/protos/variables.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1b\x61ngr/protos/variables.proto\x12\x0b\x61ngr.protos\"\x90\x01\n\x0cVariableBase\x12\r\n\x05ident\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x06region\x18\x03 \x01(\x04H\x00\x88\x01\x01\x12\x15\n\x08\x63\x61tegory\x18\x04 \x01(\tH\x01\x88\x01\x01\x12\x0f\n\x07renamed\x18\x05 \x01(\x08\x12\x0e\n\x06is_phi\x18\x06 \x01(\x08\x42\t\n\x07_regionB\x0b\n\t_category\"L\n\x11TemporaryVariable\x12\'\n\x04\x62\x61se\x18\x01 \x01(\x0b\x32\x19.angr.protos.VariableBase\x12\x0e\n\x06tmp_id\x18\x02 \x01(\r\"V\n\x10RegisterVariable\x12\'\n\x04\x62\x61se\x18\x01 \x01(\x0b\x32\x19.angr.protos.VariableBase\x12\x0b\n\x03reg\x18\x02 \x01(\r\x12\x0c\n\x04size\x18\x03 \x01(\r\"U\n\x0eMemoryVariable\x12\'\n\x04\x62\x61se\x18\x01 \x01(\x0b\x32\x19.angr.protos.VariableBase\x12\x0c\n\x04\x61\x64\x64r\x18\x02 \x01(\x04\x12\x0c\n\x04size\x18\x03 \x01(\r\"u\n\rStackVariable\x12\'\n\x04\x62\x61se\x18\x01 \x01(\x0b\x32\x19.angr.protos.VariableBase\x12\x0c\n\x04\x61\x64\x64r\x18\x02 \x01(\x04\x12\x0c\n\x04size\x18\x03 \x01(\r\x12\x0f\n\x07sp_base\x18\x04 \x01(\x08\x12\x0e\n\x06offset\x18\x05 \x01(\x05\"\x9c\x02\n\x0eVariableAccess\x12\r\n\x05ident\x18\x01 \x01(\t\x12\x12\n\nblock_addr\x18\x02 \x01(\x04\x12\x10\n\x08stmt_idx\x18\x03 \x01(\x05\x12\x10\n\x08ins_addr\x18\x04 \x01(\x04\x12\x13\n\x06offset\x18\x05 \x01(\x03H\x00\x88\x01\x01\x12\x43\n\x0b\x61\x63\x63\x65ss_type\x18\x06 \x01(\x0e\x32..angr.protos.VariableAccess.VariableAccessSort\x12\x16\n\tatom_hash\x18\x07 \x01(\rH\x01\x88\x01\x01\"8\n\x12VariableAccessSort\x12\t\n\x05WRITE\x10\x00\x12\x08\n\x04READ\x10\x01\x12\r\n\tREFERENCE\x10\x02\x42\t\n\x07_offsetB\x0c\n\n_atom_hash\"/\n\x0cVariableType\x12\r\n\x05ident\x18\x01 \x01(\t\x12\x10\n\x08var_type\x18\x02 \x01(\t\";\n\x0bVar2Unified\x12\x11\n\tvar_ident\x18\x01 \x01(\t\x12\x19\n\x11unified_var_ident\x18\x02 \x01(\t\"/\n\x07Phi2Var\x12\x11\n\tphi_ident\x18\x01 \x01(\t\x12\x11\n\tvar_ident\x18\x02 \x01(\t\"\xe6\x04\n\x17VariableManagerInternal\x12\x30\n\x08tempvars\x18\x01 \x03(\x0b\x32\x1e.angr.protos.TemporaryVariable\x12.\n\x07regvars\x18\x02 \x03(\x0b\x32\x1d.angr.protos.RegisterVariable\x12,\n\x07memvars\x18\x03 \x03(\x0b\x32\x1b.angr.protos.MemoryVariable\x12-\n\tstackvars\x18\x04 \x03(\x0b\x32\x1a.angr.protos.StackVariable\x12-\n\x08\x61\x63\x63\x65sses\x18\x05 \x03(\x0b\x32\x1b.angr.protos.VariableAccess\x12\x38\n\x10unified_tempvars\x18\x06 \x03(\x0b\x32\x1e.angr.protos.TemporaryVariable\x12\x36\n\x0funified_regvars\x18\x07 \x03(\x0b\x32\x1d.angr.protos.RegisterVariable\x12\x34\n\x0funified_memvars\x18\x08 \x03(\x0b\x32\x1b.angr.protos.MemoryVariable\x12\x35\n\x11unified_stackvars\x18\t \x03(\x0b\x32\x1a.angr.protos.StackVariable\x12-\n\x0bvar2unified\x18\n \x03(\x0b\x32\x18.angr.protos.Var2Unified\x12(\n\x05types\x18\x0b \x03(\x0b\x32\x19.angr.protos.VariableType\x12%\n\x07phi2var\x18\x0c \x03(\x0b\x32\x14.angr.protos.Phi2Varb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'angr.protos.variables_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_VARIABLEBASE']._serialized_start=45
  _globals['_VARIABLEBASE']._serialized_end=189
  _globals['_TEMPORARYVARIABLE']._serialized_start=191
  _globals['_TEMPORARYVARIABLE']._serialized_end=267
  _globals['_REGISTERVARIABLE']._serialized_start=269
  _globals['_REGISTERVARIABLE']._serialized_end=355
  _globals['_MEMORYVARIABLE']._serialized_start=357
  _globals['_MEMORYVARIABLE']._serialized_end=442
  _globals['_STACKVARIABLE']._serialized_start=444
  _globals['_STACKVARIABLE']._serialized_end=561
  _globals['_VARIABLEACCESS']._serialized_start=564
  _globals['_VARIABLEACCESS']._serialized_end=848
  _globals['_VARIABLEACCESS_VARIABLEACCESSSORT']._serialized_start=767
  _globals['_VARIABLEACCESS_VARIABLEACCESSSORT']._serialized_end=823
  _globals['_VARIABLETYPE']._serialized_start=850
  _globals['_VARIABLETYPE']._serialized_end=897
  _globals['_VAR2UNIFIED']._serialized_start=899
  _globals['_VAR2UNIFIED']._serialized_end=958
  _globals['_PHI2VAR']._serialized_start=960
  _globals['_PHI2VAR']._serialized_end=1007
  _globals['_VARIABLEMANAGERINTERNAL']._serialized_start=1010
  _globals['_VARIABLEMANAGERINTERNAL']._serialized_end=1624
# @@protoc_insertion_point(module_scope)



from __future__ import annotations
import itertools
import logging

import networkx as nx

from angr.ailment.block import Block

from .utils import bfs_list_blocks
from angr.analyses.decompiler.block_similarity import longest_ail_subseq, is_similar

_l = logging.getLogger(name=__name__)


def longest_ail_graph_subseq(block_list, graph):
    # generate a graph similarity for each pair in the provided blocks
    all_sims = [ail_graph_similarity(pair[0], pair[1], graph) for pair in itertools.combinations(block_list, 2)]

    lcs, _ = longest_ail_subseq(all_sims, graph=graph)
    return lcs


def ail_graph_similarity(block0: Block, block1: Block, graph: nx.DiGraph, only_blocks=False):
    b0_blocks = bfs_list_blocks(block0, graph)
    b1_blocks = bfs_list_blocks(block1, graph)
    similarity = []

    discontinuity_blocks = set()
    for i, b0 in enumerate(b0_blocks):
        # getting to a block with no matching index is grounds to stop cmp
        try:
            b1 = b1_blocks[i]
        except IndexError:
            break

        # if a block in the chain before did not end in LCS, don't follow it.
        if b0 in discontinuity_blocks or b1 in discontinuity_blocks:
            continue

        # SPECIAL CASE: 1
        # ┌─┐         ┌─┐
        # │A├─┐     ┌─┤A│
        # └─┘ │     │ └─┘
        #     ▼     ▼
        #    ┌──┐ ┌──┐
        #    │C'│ │C'│
        #    └┬─┘ └┬─┘
        #     │    │
        #     ▼    ▼
        #       ...
        #
        # Both similar blocks end in a block that is actually the same block. In this case, we don't
        # want to count this since we will create a N blocks long similarity that continues from C' all
        # the way to the end of the graph. This is only true if C' is not the end block of the graph.
        if b0 is b1:
            preds = list(graph.predecessors(b0))
            all_match = True
            for pred1 in preds:
                for pred2 in preds:
                    if set(graph.successors(pred1)) != set(graph.successors(pred2)):
                        all_match = False
                        break

                if not all_match:
                    break

            # CASE 1 confirmed, all edges in look the same. Now we check a special subset to see if the
            # matching node has any edges out. If this node has no edges out, we are no longer ture for 1A
            if all_match and len(list(graph.successors(b0))) != 0:
                continue

        lcs, lcs_idxs = longest_ail_subseq([b0.statements, b1.statements], graph=graph)
        if not lcs:
            break

        # verify all the blocks end in that statement or exclude its children
        for idx, b in enumerate([b0, b1]):
            if len(lcs) + lcs_idxs[idx] != len(b.statements):
                discontinuity_blocks.update(set(graph.successors(b)))

        # can output blocks only if needed
        similarity += lcs if not only_blocks else [(b0, b1)]

    return similarity


def ail_similarity_to_orig_blocks(orig_block, graph_similarity, graph):
    traversal_blocks = bfs_list_blocks(orig_block, graph)

    graph_sim = graph_similarity.copy()
    orig_blocks = []
    split_blocks = {}  # [block] = (lcs, idx)
    for block in traversal_blocks:
        if not graph_sim:
            break

        lcs, lcs_idxs = longest_ail_subseq([block.statements, graph_sim[: len(block.statements)]], graph=graph)
        if block is orig_block:
            lcs_1, lcs_idxs_1 = longest_ail_subseq([graph_sim[: len(block.statements)], block.statements], graph=graph)
            if lcs_idxs_1[1] > lcs_idxs[0]:
                lcs, lcs_idxs = lcs_1, lcs_idxs_1[::-1]

        if not lcs:
            break

        orig_blocks.append(block)

        if len(lcs) != len(block.statements):
            split_blocks[block] = (lcs, lcs_idxs[0])

        graph_sim = graph_sim[len(lcs) :]

    return orig_blocks, split_blocks


def find_block_by_similarity(block, graph, node_list=None):
    nodes = node_list if node_list else list(graph.nodes())
    similar_blocks = []
    for other_block in nodes:
        if is_similar(block, other_block, graph=graph):
            similar_blocks.append(other_block)

    if len(similar_blocks) > 1:
        _l.warning("found multiple similar blocks")

    return similar_blocks[0]



from __future__ import annotations
from typing import TYPE_CHECKING
import logging

from cle.backends.elf.variable import Variable
from cle.backends.elf.variable_type import VariableType, PointerType, ArrayType, StructType, TypedefType

from angr.sim_state import SimState
from angr.sim_type import ALL_TYPES, SimTypeNum

from .plugin import SimStatePlugin

l = logging.getLogger(name=__name__)

if TYPE_CHECKING:
    from angr.state_plugins.view import SimMemView
    from angr.sim_type import SimType


class SimDebugVariable:
    """
    A SimDebugVariable will get dynamically created when queriyng for variable in a state with the
    SimDebugVariablePlugin. It features a link to the state, an address and a type.
    """

    def __init__(self, state: SimState, addr, var_type: VariableType):
        self.state = state
        self.addr = addr
        self.type = var_type

    @staticmethod
    def from_cle_variable(state: SimState, cle_variable: Variable, dwarf_cfa) -> SimDebugVariable:
        addr = cle_variable.rebased_addr_from_cfa(dwarf_cfa)
        var_type = cle_variable.type
        return SimDebugVariable(state, addr, var_type)

    @property
    def mem_untyped(self) -> SimMemView:
        if self.addr is None:
            raise Exception("Cannot view a variable without an address")
        return self.state.mem[self.addr]

    @property
    def mem(self) -> SimMemView:
        if isinstance(self.type, TypedefType):
            unpacked = SimDebugVariable(self.state, self.addr, self.type.type)
            return unpacked.mem
        if self.type is None or self.type.byte_size is None:
            return self.mem_untyped

        arch = self.state.arch
        size = self.type.byte_size * arch.byte_width
        name = self.type.name
        if name in ALL_TYPES:
            sim_type = ALL_TYPES[name].with_arch(arch)
            assert size == sim_type.size
        else:
            # FIXME A lot more types are supported by angr that are not in ALL_TYPES (structs, arrays, pointers)
            # Use a fallback type
            sim_type = SimTypeNum(size, signed=False, label=name)
        return self.mem_untyped.with_type(sim_type)

    # methods and properties equivalent to SimMemView

    @property
    def string(self) -> SimMemView:
        first_char = self.deref
        # first char should have some char type (could be checked here)
        return first_char.mem_untyped.string

    def with_type(self, sim_type: SimType) -> SimMemView:
        return self.mem_untyped.with_type(sim_type)

    @property
    def resolvable(self):
        return self.mem.resolvable

    @property
    def resolved(self):
        return self.mem.resolved

    @property
    def concrete(self):
        return self.mem.concrete

    def store(self, value):
        return self.mem.store(value)

    def __getitem__(self, i):
        if isinstance(i, int):
            return self.array(i)
        if isinstance(i, str):
            return self.member(i)
        raise KeyError

    @property
    def deref(self) -> SimDebugVariable:
        # dereferincing is equivalent to getting the first array element
        return self.array(0)

    def array(self, i) -> SimDebugVariable:
        if isinstance(self.type, TypedefType):
            unpacked = SimDebugVariable(self.state, self.addr, self.type.type)
            return unpacked.array(i)
        if isinstance(self.type, ArrayType):
            # an array already addresses its first element
            addr = self.addr
            el_type = self.type.element_type
        elif isinstance(self.type, PointerType):
            if self.addr is None:
                addr = None
            else:
                addr = self.state.memory.load(self.addr, self.state.arch.bytes, endness=self.state.arch.memory_endness)
            el_type = self.type.referenced_type
        else:
            raise Exception(f"{self.type} object cannot be dereferenced")

        if i == 0:
            new_addr = addr
        elif addr is None or el_type is None or el_type.byte_size is None:
            new_addr = None
        else:
            new_addr = addr + i * el_type.byte_size
        return SimDebugVariable(self.state, new_addr, el_type)

    def member(self, member_name: str) -> SimDebugVariable:
        if isinstance(self.type, TypedefType):
            unpacked = SimDebugVariable(self.state, self.addr, self.type.type)
            return unpacked.member(member_name)
        if isinstance(self.type, StructType):
            member = self.type[member_name]
            addr = None if self.addr is None else self.addr + member.addr_offset
            return SimDebugVariable(self.state, addr, member.type)

        raise Exception(f"{self.type} object has no members")


class SimDebugVariablePlugin(SimStatePlugin):
    """
    This is the plugin you'll use to interact with (global/local) program variables.
    These variables have a name and a visibility scope which depends on the pc address of the state.
    With this plugin, you can access/modify the value of such variable or find its memory address.
    For creating program variables, or for importing them from cle, see the knowledge plugin debug_variables.
    Run ``p.kb.dvars.load_from_dwarf()`` before using this plugin.

    Example:
        >>> p = angr.Project("various_variables", load_debug_info=True)
        >>> p.kb.dvars.load_from_dwarf()
        >>> state =  # navigate to the state you want
        >>> state.dvars.get_variable("pointer2").deref.mem
        <int (32 bits) <BV32 0x1> at 0x404020>
    """

    def get_variable(self, var_name: str) -> SimDebugVariable:
        """
        Returns the visible variable (if any) with name ``var_name`` based on the current ``state.ip``.
        """
        kb = self.state.project.kb
        cle_var = kb.dvars[var_name][self.state.ip]
        if cle_var:
            return SimDebugVariable.from_cle_variable(self.state, cle_var, self.dwarf_cfa)
        return None

    def __getitem__(self, var_name: str) -> SimDebugVariable:
        return self.get_variable(var_name)

    # DWARF cfa
    @property
    def dwarf_cfa(self):
        """
        Returns the current cfa computation. Set this property to the correct value if needed.
        """
        try:
            return self._dwarf_cfa
        except AttributeError:
            return self.dwarf_cfa_approx

    @dwarf_cfa.setter
    def dwarf_cfa(self, new_val):
        self._dwarf_cfa = new_val

    @property
    def dwarf_cfa_approx(self):
        # FIXME This is only an approximation!
        if self.state.arch.name == "AMD64":
            return self.state.regs.rbp + 16
        if self.state.arch.name == "X86":
            return self.state.regs.ebp + 8
        return 0


SimState.register_default("dvars", SimDebugVariablePlugin)



# Util functions that are mostly rewrite of the original code in redare2
from __future__ import annotations
import struct


def read_short(file_obj) -> int:
    return struct.unpack(">H", file_obj.read(2))[0]


def read_word(file_obj) -> int:
    return struct.unpack(">I", file_obj.read(4))[0]


def read_max_2_bytes(file_obj) -> int:
    b = file_obj.read(1)[0]
    if b & 0x80 != 0x80:
        return b
    return ((b & 0x7F) << 8) + file_obj.read(1)[0]


def read_multiple_bytes(file_obj) -> int:
    b = file_obj.read(1)[0]
    if b & 0x80 != 0x80:
        return b
    if b & 0xC0 != 0xC0:
        return ((b & 0x7F) << 8) + file_obj.read(1)[0]
    if b & 0xE0 != 0xE0:
        b = ((b & 0x3F) << 24) + (file_obj.read(1)[0] << 16)
        b += read_short(file_obj)
        return b
    return read_word(file_obj)



# pylint:disable=arguments-renamed,too-many-boolean-expressions
from __future__ import annotations
from typing import Any

import angr.ailment as ailment
from angr.ailment.expression import Op

from angr.analyses.decompiler.structuring.structurer_nodes import ConditionNode
from angr.analyses.decompiler.utils import (
    structured_node_is_simple_return,
    structured_node_is_simple_return_strict,
    sequence_to_statements,
    structured_node_has_multi_predecessors,
)
from angr.analyses.decompiler.sequence_walker import SequenceWalker
from .optimization_pass import SequenceOptimizationPass, OptimizationPassStage


class FlipBooleanWalker(SequenceWalker):
    """
    Walks a SequenceNode and handles every sequence.
    Uses the flip_size to determine when to flip the condition on large if-statement bodies.
    """

    def __init__(self, graph, flip_size=9, last_node=None):
        super().__init__()
        self._graph = graph
        self._last_node = last_node
        self._flip_size = flip_size

    def _handle_Sequence(self, seq_node, **kwargs):
        # Type 1:
        # if (cond) { ... }  else  { return; }   -->   if (!cond) { return; } else { ... }
        #
        # Type 2:
        # if (cond) { ... } return;    -->    if (!cond) return; ...; return;
        type1_condition_nodes = [node for node in seq_node.nodes if isinstance(node, ConditionNode) and node.false_node]
        type2_condition_nodes: list[tuple[int, ConditionNode, Any]] = []

        if len(seq_node.nodes) >= 2:
            idx = len(seq_node.nodes) - 2
            node = seq_node.nodes[idx]
            if (
                isinstance(node, ConditionNode)
                and node.true_node is not None
                and node.false_node is None
                and idx < len(seq_node.nodes) - 1
                and structured_node_is_simple_return_strict(seq_node.nodes[idx + 1])
                and node not in type1_condition_nodes
            ):
                # Type 2: Special Filter:
                # consider code that looks like the following:
                # {if (cond) {LABEL: ... } return;}; goto LABEL;
                #
                # if we were to do the normal flip, this happens:
                # {if (!cond) return; LABEL: ...}; goto LABEL;
                #
                # This is incorrect because we've now created an infinite loop in the event that cond is false,
                # which is not what the original code was. The gist here is that you can't ever flip these cases
                # in the presence of more than one incoming edge to `...` region.
                #
                # To eliminate this illegal case, we simply need to find all the condition nodes of the above structure
                # that have multiple incoming edges to the `...` region.
                illegal_flip = structured_node_has_multi_predecessors(node.true_node, self._graph)
                if not illegal_flip:
                    type2_condition_nodes.append((idx, node, seq_node.nodes[idx + 1]))

        for node in type1_condition_nodes:
            if isinstance(node.condition, Op) and structured_node_is_simple_return(node.false_node, self._graph):
                node.condition = ailment.expression.negate(node.condition)
                node.true_node, node.false_node = node.false_node, node.true_node

        for idx, cond_node, successor in type2_condition_nodes:
            # there are two possibilities when you might want to flip the condition and move the return statement:
            # 1. This if-stmt if found somewhere in the middle of the function
            # 2. This if-stmt is pretty large, but still ends in a return outside of the if-stmt
            if (successor is not self._last_node) or (
                len(sequence_to_statements(cond_node.true_node)) >= self._flip_size
            ):
                cond_node.condition = ailment.expression.negate(cond_node.condition)
                seq_node.nodes[idx + 1] = cond_node.true_node
                cond_node.true_node = successor
                seq_node.nodes.insert(idx + 2, successor)

        return super()._handle_Sequence(seq_node, **kwargs)


class FlipBooleanCmp(SequenceOptimizationPass):
    """
    In the scenario in which a false node has no apparent successors, flip the condition on that if-stmt.
    This is only useful when StructuredCodeGenerator has simplify_else_scopes enabled, as this will allow the
    flipped if-stmt to remove the redundant else.
    """

    ARCHES = None
    PLATFORMS = None
    STAGE = OptimizationPassStage.AFTER_STRUCTURING
    NAME = "Flip small ret booleans"
    DESCRIPTION = "When false node has no successors, flip condition so else scope can be simplified later"

    def __init__(self, func, flip_size=9, **kwargs):
        super().__init__(func, **kwargs)
        self._graph = kwargs.get("graph")
        self._flip_size = flip_size
        self.analyze()

    def _check(self):
        return bool(self.seq.nodes), None

    def _analyze(self, cache=None):
        walker = FlipBooleanWalker(self._graph, last_node=self.seq.nodes[-1], flip_size=self._flip_size)
        walker.walk(self.seq)
        self.out_seq = self.seq



from __future__ import annotations
import claripy

import angr


class __p__fmode(angr.SimProcedure):
    def run(self):
        return self.project.simos.fmode_ptr


class _get_fmode(angr.SimProcedure):
    def run(self, outptr):
        if self.state.solver.is_true(outptr == 0):
            return 22
        fmode = self.state.mem[self.project.simos.fmode_ptr].int.resolved
        self.state.mem[outptr].int = fmode
        return 0


class _set_fmode(angr.SimProcedure):
    def run(self, val):
        if not self.state.solver.is_true(claripy.Or(val == 0x4000, val == 0x8000)):
            return 22
        self.state.mem[self.project.simos.fmode_ptr].int = val
        return 0


class __p__commode(angr.SimProcedure):
    def run(self):
        return self.project.simos.commode_ptr



# pylint:disable=arguments-differ
from __future__ import annotations
from typing import Literal, Any
from collections import defaultdict

from angr.ailment.expression import Const, StackBaseOffset, VirtualVariable, BinaryOp
from angr.ailment.statement import Call, Assignment, Store, Statement
from angr.ailment.utils import is_none_or_likeable
from angr import SIM_LIBRARIES
from .base import PeepholeOptimizationStmtBase


class ConstAssignmentInfo:
    """
    Holds information about a constant assignment.
    """

    __slots__ = ("base", "bits", "count", "kind", "offset", "value")

    def __init__(
        self, kind: Literal["stack", "heap", "global"], base: Any, offset: int, value: int, bits: int, count: int = 1
    ):
        self.kind: Literal["stack", "heap", "global"] = kind
        self.base = base
        self.offset = offset
        self.value = value
        self.bits = bits
        self.count = count

    def __repr__(self):
        base_str = str(self.base) if self.base is not None else ""
        offset_str = f"+{self.offset}" if self.offset >= 0 else f"{self.offset}"
        return (
            f"ConstAssignmentInfo({self.kind}@{base_str}{offset_str}, {self.value}|{self.bits}bits, count={self.count})"
        )


class InlinedMemset(PeepholeOptimizationStmtBase):
    """
    Simplifies inlined memory setting logic into calls to memset
    """

    MIN_ASSIGNMENTS = 2

    __slots__ = ()

    NAME = "Simplifying inlined memset"
    stmt_classes = (Assignment, Store)

    def optimize(self, stmt: Assignment | Store, stmt_idx: int | None = None, block=None, **kwargs):
        d = self._parse_const_assignment(stmt)
        if d is None:
            return None

        assert block is not None and self.project is not None

        stmt_idx = -1 if stmt_idx is None else stmt_idx
        info: list[tuple[int, ConstAssignmentInfo]] = [(stmt_idx, d)]
        # search forward for constant stores
        i = stmt_idx + 1
        while i < len(block.statements):
            d = self._parse_const_assignment(block.statements[i])
            if d is None:
                break
            info.append((i, d))
            i += 1

        # analyze the collected info
        info_by_kind = defaultdict(list)
        for idx, d in info:
            info_by_kind[d.kind].append((idx, d))

        result = None
        optimized = False
        for _kind, lst in info_by_kind.items():
            if len(lst) <= 1:
                continue
            candidates = self._find_memset_candidates(lst)
            for start_stmt_idx, end_stmt_idx, candidate in candidates:
                if end_stmt_idx - start_stmt_idx <= self.MIN_ASSIGNMENTS:
                    continue

                for i in range(start_stmt_idx, end_stmt_idx):
                    block.statements[i] = None  # remove these statements
                # create a new memset call
                base_expr = None

                match candidate.kind:
                    case "stack":
                        base_expr = StackBaseOffset(None, self.project.arch.bits, candidate.offset)
                    case "global":
                        base_expr = Const(None, None, candidate.offset, self.project.arch.bits)
                    case "heap":
                        base_expr = BinaryOp(
                            None,
                            "Add",
                            [
                                candidate.base,
                                Const(None, None, candidate.offset, self.project.arch.bits),
                            ],
                            False,
                            bits=self.project.arch.bits,
                        )

                if base_expr is None:
                    continue

                assert self.project is not None
                call_stmt = Call(
                    stmt.idx,
                    "memset",
                    args=[
                        base_expr,
                        Const(None, None, candidate.value, 8),
                        Const(None, None, candidate.count, self.project.arch.bits),
                    ],
                    prototype=SIM_LIBRARIES["libc.so"][0].get_prototype("memcpy"),
                    **stmt.tags,
                )
                if start_stmt_idx == stmt_idx:
                    result = call_stmt
                block.statements[start_stmt_idx] = call_stmt
                optimized = True

        if optimized:
            # some optimization has been made
            block.statements = [ss for ss in block.statements if ss is not None]

        return result

    @staticmethod
    def _parse_const_assignment(
        stmt: Statement,
    ) -> ConstAssignmentInfo | None:
        """
        Parse a constant assignment statement to see if it matches the pattern of a memset.

        :param stmt: The statement to parse.
        :return: A tuple of ("stack" | "heap" | "global", base, offset, value, bits) if it matches the pattern, else
                 None.
        """
        if isinstance(stmt, Assignment) and isinstance(stmt.src, Const) and stmt.src.is_int:
            dst = stmt.dst
            if isinstance(dst, VirtualVariable) and dst.was_stack:
                return ConstAssignmentInfo("stack", None, dst.stack_offset, stmt.src.value_int, stmt.src.bits)
        elif isinstance(stmt, Store) and isinstance(stmt.data, Const) and stmt.data.is_int:
            if isinstance(stmt.addr, StackBaseOffset):
                return ConstAssignmentInfo("stack", None, stmt.addr.offset, stmt.data.value_int, stmt.data.bits)
            if isinstance(stmt.addr, Const):
                return ConstAssignmentInfo("global", stmt.addr.value_int, 0, stmt.data.value_int, stmt.data.bits)
            if isinstance(stmt.addr, BinaryOp) and stmt.addr.op in {"Add", "Sub"}:
                base = stmt.addr.operands[0]
                offset_expr = stmt.addr.operands[1]
                if isinstance(offset_expr, Const):
                    offset = offset_expr.value_int if stmt.addr.op == "Add" else -offset_expr.value_int
                    return ConstAssignmentInfo("heap", base, offset, stmt.data.value_int, stmt.data.bits)
            else:
                base = stmt.addr
                return ConstAssignmentInfo("heap", base, 0, stmt.data.value_int, stmt.data.bits)
        return None

    @staticmethod
    def _find_memset_candidates(
        info_lst: list[tuple[int, ConstAssignmentInfo]],
    ) -> list[tuple[int, int, ConstAssignmentInfo]]:
        """
        Analyze a list of ConstAssignmentInfo to find candidates for memset.

        :param info_lst: The list of ConstAssignmentInfo.
        :return: A tuple of (base, start_offset, store_size, value) if they are candidates, else None.
        """
        if len(info_lst) < 2:
            return []

        candidates = []

        info_lst = sorted(info_lst, key=lambda x: x[0])
        start = 0
        while start < len(info_lst) - 1:
            # assignments must be in consecutive statements
            j = start
            while j < len(info_lst):
                if j - start != info_lst[j][0] - info_lst[start][0]:
                    break
                j += 1

            if j - start < 2:
                start += 1
                continue
            end = j

            # assignments from start to j - 1 must cover a continuous memory region
            base = info_lst[start][1].base
            regions = {}
            byte_values = set()
            has_same_base = True

            for j in range(start, end):
                ith_base = info_lst[j][1].base

                if not is_none_or_likeable(ith_base, base):
                    has_same_base = False
                    break
                regions[info_lst[j][1].offset] = info_lst[j][1].bits
                byte_values |= InlinedMemset._int_to_bytes(info_lst[j][1].value, info_lst[j][1].bits)

            if not has_same_base:
                end = j
                if end - start < 2:
                    start += 1
                    continue

            # must have the same byte value
            if len(byte_values) != 1:
                start += 1
                continue

            # last check: are regions continuous?
            next_region_offset = None
            start_region_offset = min(regions)
            has_gap = False
            for region_offset in sorted(regions):
                region_bits = regions[region_offset]
                if next_region_offset is not None and region_offset != next_region_offset:
                    has_gap = True
                    break
                next_region_offset = region_offset + region_bits // 8

            if next_region_offset is None or has_gap:
                start += 1
                continue

            candidate = ConstAssignmentInfo(
                info_lst[start][1].kind,
                base,
                start_region_offset,
                next(iter(byte_values)),
                8,
                next_region_offset - start_region_offset,
            )
            candidates.append(
                (info_lst[start][0], info_lst[end][0] if end < len(info_lst) else info_lst[end - 1][0] + 1, candidate)
            )
            start = end

        return candidates

    @staticmethod
    def _int_to_bytes(value: int, bits: int) -> set[int]:
        """
        Convert an integer value to a set of byte values.

        :param value: The integer value.
        :param bits: The number of bits in the integer.
        :return: A set of byte values.
        """
        byte_values = set()
        num_bytes = bits // 8
        for i in range(num_bytes):
            byte = (value >> (i * 8)) & 0xFF
            byte_values.add(byte)
        return byte_values



from __future__ import annotations
import logging

import claripy

import angr

l = logging.getLogger(name=__name__)


class memcmp(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, s1_addr, s2_addr, n):
        max_memcmp_size = self.state.libc.max_buffer_size

        definite_size = self.state.solver.min_int(n)
        conditional_s1_start = s1_addr + definite_size
        conditional_s2_start = s2_addr + definite_size
        conditional_size = int(max(max_memcmp_size - definite_size, 0)) if self.state.solver.symbolic(n) else 0

        l.debug("Definite size %s and conditional size: %s", definite_size, conditional_size)

        int_bits = self.arch.sizeof["int"]
        if definite_size > 0:
            s1_part = self.state.memory.load(s1_addr, definite_size, endness="Iend_BE")
            s2_part = self.state.memory.load(s2_addr, definite_size, endness="Iend_BE")
            cases = [
                [s1_part == s2_part, claripy.BVV(0, int_bits)],
                [claripy.ULT(s1_part, s2_part), claripy.BVV(-1, int_bits)],
                [claripy.UGT(s1_part, s2_part), claripy.BVV(1, int_bits)],
            ]
            definite_answer = claripy.ite_cases(cases, 2)
            constraint = claripy.Or(*[c for c, _ in cases])
            self.state.add_constraints(constraint)

            l.debug("Created definite answer: %s", definite_answer)
            l.debug("Created constraint: %s", constraint)
            l.debug("... crom cases: %s", cases)
        else:
            definite_answer = claripy.BVV(0, int_bits)

        if not self.state.solver.symbolic(definite_answer) and self.state.solver.eval(definite_answer) != 0:
            return definite_answer

        if conditional_size > 0:
            s1_all = self.state.memory.load(conditional_s1_start, conditional_size, endness="Iend_BE")
            s2_all = self.state.memory.load(conditional_s2_start, conditional_size, endness="Iend_BE")
            conditional_rets = {0: definite_answer}

            for byte, bit in zip(range(conditional_size), range(conditional_size * 8, 0, -8)):
                s1_part = s1_all[conditional_size * 8 - 1 : bit - 8]
                s2_part = s2_all[conditional_size * 8 - 1 : bit - 8]
                cases = [
                    [s1_part == s2_part, claripy.BVV(0, int_bits)],
                    [claripy.ULT(s1_part, s2_part), claripy.BVV(-1, int_bits)],
                    [claripy.UGT(s1_part, s2_part), claripy.BVV(1, int_bits)],
                ]
                conditional_rets[byte + 1] = claripy.ite_cases(cases, 0)
                self.state.add_constraints(claripy.Or(*[c for c, _ in cases]))

            ret_expr = claripy.If(
                definite_answer == 0,
                claripy.ite_dict(n - definite_size, conditional_rets, 2),
                definite_answer,
            )
            self.state.add_constraints(claripy.Or(*[n - definite_size == c for c in conditional_rets]))
            return ret_expr
        return definite_answer



from __future__ import annotations
import random
from angr.analyses.reaching_definitions.function_handler import FunctionCallDataUnwrapped, FunctionHandler
from angr.analyses.reaching_definitions.function_handler_library.stdio import StdinAtom, StdoutAtom
from angr.analyses.reaching_definitions.rd_state import ReachingDefinitionsState
from angr.knowledge_plugins.key_definitions.atoms import Atom

# pylint: disable=no-self-use,missing-class-docstring,unused-argument


class FDAtom(Atom):
    def __init__(self, fd: int | None, source: str, size: int = 1):
        self.source = source
        self.fd = fd
        self.nonce = random.randint(0, 999999999999)
        super().__init__(size)

    def _identity(self):
        if self.fd is not None:
            return (self.fd,)
        return (self.nonce,)


class LibcUnistdHandlers(FunctionHandler):
    @FunctionCallDataUnwrapped.decorate
    def handle_impl_read(self, state: ReachingDefinitionsState, data: FunctionCallDataUnwrapped):
        size = state.get_concrete_value(data.args_atoms[2]) or 1
        dst_atom = state.deref(data.args_atoms[1], size)
        real_fd = state.get_concrete_value(data.args_atoms[0])

        fd_atom = StdinAtom(data.function.name, size) if real_fd == 0 else FDAtom(real_fd, data.function.name, size)
        buf_data = state.top(size * 8) if size is not None else state.top(state.arch.bits)

        data.depends(dst_atom, fd_atom, value=buf_data)
        # Model return value - number of bytes read (could be 0 to size, or -1 for error)
        ret_val = state.top(state.arch.bits)  # Unknown number of bytes read
        data.depends(data.ret_atoms, fd_atom, value=ret_val)

    handle_impl_recv = handle_impl_recvfrom = handle_impl_read

    @FunctionCallDataUnwrapped.decorate
    def handle_impl_write(self, state: ReachingDefinitionsState, data: FunctionCallDataUnwrapped):
        size = state.get_concrete_value(data.args_atoms[2]) or 1
        src_atom = state.deref(data.args_atoms[1], size)
        data.depends(StdoutAtom(data.function.name, size), src_atom, value=state.get_values(src_atom))

        # Model return value - number of bytes written (could be 0 to size, or -1 for error)
        ret_val = state.top(state.arch.bits)  # Unknown number of bytes written
        data.depends(data.ret_atoms, value=ret_val)

    handle_impl_send = handle_impl_write



# pylint:disable=missing-class-docstring
from __future__ import annotations
from enum import Enum


class FlirtArch(int, Enum):
    ARCH_386 = 0  # Intel 80x86
    ARCH_Z80 = 1  # 8085, Z80
    ARCH_I860 = 2  # Intel 860
    ARCH_8051 = 3  # 8051
    ARCH_TMS = 4  # Texas Instruments TMS320C5x
    ARCH_6502 = 5  # 6502
    ARCH_PDP = 6  # PDP11
    ARCH_68K = 7  # Motorola 680x0
    ARCH_JAVA = 8  # Java
    ARCH_6800 = 9  # Motorola 68xx
    ARCH_ST7 = 10  # SGS-Thomson ST7
    ARCH_MC6812 = 11  # Motorola 68HC12
    ARCH_MIPS = 12  # MIPS
    ARCH_ARM = 13  # Advanced RISC Machines
    ARCH_TMSC6 = 14  # Texas Instruments TMS320C6x
    ARCH_PPC = 15  # PowerPC
    ARCH_80196 = 16  # Intel 80196
    ARCH_Z8 = 17  # Z8
    ARCH_SH = 18  # Renesas (formerly Hitachi) SuperH
    ARCH_NET = 19  # Microsoft Visual Studio.Net
    ARCH_AVR = 20  # Atmel 8-bit RISC processor(s)
    ARCH_H8 = 21  # Hitachi H8/300, H8/2000
    ARCH_PIC = 22  # Microchip's PIC
    ARCH_SPARC = 23  # SPARC
    ARCH_ALPHA = 24  # DEC Alpha
    ARCH_HPPA = 25  # Hewlett-Packard PA-RISC
    ARCH_H8500 = 26  # Hitachi H8/500
    ARCH_TRICORE = 27  # Tasking Tricore
    ARCH_DSP56K = 28  # Motorola DSP5600x
    ARCH_C166 = 29  # Siemens C166 family
    ARCH_ST20 = 30  # SGS-Thomson ST20
    ARCH_IA64 = 31  # Intel Itanium IA64
    ARCH_I960 = 32  # Intel 960
    ARCH_F2MC = 33  # Fujitsu F2MC-16
    ARCH_TMS320C54 = 34  # Texas Instruments TMS320C54xx
    ARCH_TMS320C55 = 35  # Texas Instruments TMS320C55xx
    ARCH_TRIMEDIA = 36  # Trimedia
    ARCH_M32R = 37  # Mitsubishi 32bit RISC
    ARCH_NEC_78K0 = 38  # NEC 78K0
    ARCH_NEC_78K0S = 39  # NEC 78K0S
    ARCH_M740 = 40  # Mitsubishi 8bit
    ARCH_M7700 = 41  # Mitsubishi 16bit
    ARCH_ST9 = 42  # ST9+
    ARCH_FR = 43  # Fujitsu FR Family
    ARCH_MC6816 = 44  # Motorola 68HC16
    ARCH_M7900 = 45  # Mitsubishi 7900
    ARCH_TMS320C3 = 46  # Texas Instruments TMS320C3
    ARCH_KR1878 = 47  # Angstrem KR1878
    ARCH_AD218X = 48  # Analog Devices ADSP 218X
    ARCH_OAKDSP = 49  # Atmel OAK DSP
    ARCH_TLCS900 = 50  # Toshiba TLCS-900
    ARCH_C39 = 51  # Rockwell C39
    ARCH_CR16 = 52  # NSC CR16
    ARCH_MN102L00 = 53  # Panasonic MN10200
    ARCH_TMS320C1X = 54  # Texas Instruments TMS320C1x
    ARCH_NEC_V850X = 55  # NEC V850 and V850ES/E1/E2
    ARCH_SCR_ADPT = 56  # Processor module adapter for processor modules written in scripting languages
    ARCH_EBC = 57  # EFI Bytecode
    ARCH_MSP430 = 58  # Texas Instruments MSP430
    ARCH_SPU = 59  # Cell Broadband Engine Synergistic Processor Unit
    ARCH_DALVIK = 60  # Android Dalvik Virtual Machine


FLIRT_ARCH_TO_ARCHNAME: dict[int, dict[int, str]] = {
    FlirtArch.ARCH_386: {32: "X86", 64: "AMD64"},
    FlirtArch.ARCH_MIPS: {32: "MIPS32", 64: "MIPS64"},
    FlirtArch.ARCH_ARM: {32: "ARM", 64: "AARCH64"},
    FlirtArch.ARCH_PPC: {32: "PPC32", 64: "PPC64"},
    FlirtArch.ARCH_SPARC: {32: "SPARC32", 64: "SPARC64"},
}


class FlirtFileType(int, Enum):
    FILE_DOS_EXE_OLD = 0x00000001
    FILE_DOS_COM_OLD = 0x00000002
    FILE_BIN = 0x00000004
    FILE_DOSDRV = 0x00000008
    FILE_NE = 0x00000010
    FILE_INTELHEX = 0x00000020
    FILE_MOSHEX = 0x00000040
    FILE_LX = 0x00000080
    FILE_LE = 0x00000100
    FILE_NLM = 0x00000200
    FILE_COFF = 0x00000400
    FILE_PE = 0x00000800
    FILE_OMF = 0x00001000
    FILE_SREC = 0x00002000
    FILE_ZIP = 0x00004000
    FILE_OMFLIB = 0x00008000
    FILE_AR = 0x00010000
    FILE_LOADER = 0x00020000
    FILE_ELF = 0x00040000
    FILE_W32RUN = 0x00080000
    FILE_AOUT = 0x00100000
    FILE_PILOT = 0x00200000
    FILE_DOS_EXE = 0x00400000
    FILE_DOS_COM = 0x00800000
    FILE_AIXAR = 0x01000000


class FlirtOSType(int, Enum):
    """
    Actually no longer used in IDA.
    """

    OS_MSDOS = 0x01
    OS_WIN = 0x02
    OS_OS2 = 0x04
    OS_NETWARE = 0x08
    OS_UNIX = 0x10
    OS_OTHER = 0x20


FLIRT_OS_TO_OSNAME = {
    FlirtOSType.OS_MSDOS: "MSDOS",
    FlirtOSType.OS_WIN: "Win32",
    FlirtOSType.OS_OS2: "OS/2",
    FlirtOSType.OS_UNIX: "Linux",
    FlirtOSType.OS_OTHER: "Other",
}


class FlirtAppType(int, Enum):
    APP_CONSOLE = 0x0001
    APP_GRAPHICS = 0x0002
    APP_EXE = 0x0004
    APP_DLL = 0x0008
    APP_DRV = 0x0010
    APP_SINGLE_THREADED = 0x0020
    APP_MULTI_THREADED = 0x0040
    APP_16_BIT = 0x0080
    APP_32_BIT = 0x0100
    APP_64_BIT = 0x0200


class FlirtFeatureFlag(int, Enum):
    FEATURE_STARTUP = 0x1
    FEATURE_CTYPE_CRC = 0x2
    FEATURE_2BYTE_CTYPE = 0x4
    FEATURE_ALT_CTYPE_CRC = 0x8
    FEATURE_COMPRESSED = 0x10


class FlirtParseFlag(int, Enum):
    PARSE_MORE_PUBLIC_NAMES = 0x1
    PARSE_READ_TAIL_BYTES = 0x2
    PARSE_READ_REFERENCED_FUNCTIONS = 0x4
    PARSE_MORE_MODULES_WITH_SAME_CRC = 0x8
    PARSE_MORE_MODULES = 0x10


class FlirtFunctionFlag(int, Enum):
    FUNCTION_LOCAL = 0x2
    FUNCTION_UNRESOLVED_COLLISION = 0x8



from __future__ import annotations
from typing import TYPE_CHECKING, cast

from collections.abc import Iterator
from dataclasses import dataclass
import logging

from angr.project import Project
from angr.knowledge_base import KnowledgeBase
from angr.knowledge_plugins.functions.function import Function
from angr.knowledge_plugins.key_definitions import DerefSize
from angr.knowledge_plugins.key_definitions.constants import ObservationPointType
from angr.knowledge_plugins.key_definitions.atoms import MemoryLocation
from angr.sim_variable import SimMemoryVariable

if TYPE_CHECKING:
    from angr.analyses.reaching_definitions import (
        ReachingDefinitionsAnalysis,
        FunctionCallRelationships,
    )


log = logging.getLogger(__name__)


@dataclass
class APIObfuscationType2:
    """
    API Obfuscation Type 2 result.
    """

    resolved_func_name: str
    resolved_func_ptr: MemoryLocation
    resolved_in: Function
    resolved_by: Function


class APIObfuscationType2Finder:
    """
    Finds global function pointers initialized by calls to dlsym/GetProcAddress and names
    them accordingly.
    """

    results: list[APIObfuscationType2]

    def __init__(self, project: Project, variable_kb: KnowledgeBase | None = None):
        self.project = project
        self.variable_kb = variable_kb or self.project.kb
        self.results = []

    def analyze(self) -> list[APIObfuscationType2]:
        self.results = []
        for caller, callee in self._get_candidates():
            rda = self.project.analyses.ReachingDefinitions(caller, observe_all=True)
            for info in rda.callsites_to(callee):
                self._process_callsite(caller, callee, rda, info)
        self._mark_globals()
        return self.results

    def _get_candidates(self) -> Iterator[tuple[Function, Function]]:
        """
        Returns a tuple of (caller, callee) where callee is GetProcAddress/dlsym.
        """
        targets = ["GetProcAddress"] if self.project.simos.name == "Win32" else ["dlsym", "dlvsym"]
        for callee in self.project.kb.functions.values():
            if callee.name not in targets:
                continue
            for caller_addr in self.project.kb.callgraph.predecessors(callee.addr):
                caller = self.project.kb.functions[caller_addr]
                yield (caller, callee)

    def _process_callsite(
        self,
        caller: Function,
        callee: Function,
        rda: ReachingDefinitionsAnalysis,
        callsite_info: FunctionCallRelationships,
    ) -> None:
        """
        Process a resolver function callsite looking for function name concrete string argument.
        """
        func_name_arg_idx = 1
        if len(callsite_info.args_defns) <= func_name_arg_idx:
            return

        log.debug("Examining call to %r from %r at %r", callee, caller, callsite_info.callsite.ins_addr)
        ld = rda.model.get_observation_by_insn(callsite_info.callsite, ObservationPointType.OP_BEFORE)
        if ld is None:
            return

        # Attempt resolving static function name from callsite
        string_atom = ld.deref(callsite_info.args_defns[func_name_arg_idx], DerefSize.NULL_TERMINATE)
        result = ld.get_concrete_value(string_atom, cast_to=bytes)
        if result is None:
            log.debug("...Failed to resolve a function name")
            return

        try:
            func_name = result.rstrip(b"\x00").decode("utf-8")
            log.debug("...Resolved concrete function name: %s", func_name)
        except UnicodeDecodeError:
            log.debug("...Failed to decode utf-8 function name")
            return

        # Examine successor definitions to find where the function pointer is written
        for successor in rda.dep_graph.find_all_successors(callsite_info.ret_defns):
            if not (
                isinstance(successor.atom, MemoryLocation)
                and isinstance(successor.atom.addr, int)
                and successor.atom.size == self.project.arch.bytes
            ):
                continue

            ptr = successor.atom
            ptr_addr: int = cast(int, ptr.addr)
            log.debug("...Found function pointer %r", ptr)

            sym = self.project.loader.find_symbol(ptr_addr)
            if sym is not None:
                log.debug("...Already have pointer symbol: %r. Skipping.", sym)
                continue
            if ptr_addr in self.project.kb.labels:
                log.debug("...Already have pointer label. Skipping.")
                continue
            sec = self.project.loader.find_section_containing(ptr_addr)
            if not sec or not sec.is_writable:
                log.debug("...Bogus section. Skipping.")
                continue

            self.results.append(
                APIObfuscationType2(
                    resolved_func_name=func_name,
                    resolved_func_ptr=ptr,
                    resolved_in=caller,
                    resolved_by=callee,
                )
            )

    def _mark_globals(self):
        """
        Create function pointer labels/variables.
        """
        for result in self.results:
            # Create a label
            lbl = self.project.kb.labels.get_unique_label(f"p_{result.resolved_func_name}")
            self.project.kb.labels[result.resolved_func_ptr.addr] = lbl
            log.debug("...Created label %s for address %x", lbl, result.resolved_func_ptr.addr)

            # Create a variable
            global_variables = self.variable_kb.variables["global"]
            variables = global_variables.get_global_variables(result.resolved_func_ptr.addr)
            if not variables:
                ident = global_variables.next_variable_ident("global")
                var = SimMemoryVariable(
                    result.resolved_func_ptr.addr, result.resolved_func_ptr.size, name=lbl, ident=ident
                )
                global_variables.set_variable("global", result.resolved_func_ptr.addr, var)
                log.debug("...Created variable %r", var)
            elif len(variables) == 1:
                (var,) = variables
                log.debug("...Renaming variable %r -> %s", var, lbl)
                var.name = lbl

            self.project.kb.obfuscations.type2_deobfuscated_apis[result.resolved_func_ptr.addr] = (
                result.resolved_func_name
            )
